# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import datetime
import os
from pathlib import Path
from time import time
from typing import Any, Optional, TypeVar
from uuid import uuid4

import h5py
import importlib_metadata
import numpy as np

from acconeer.exptool.a121._core import utils
from acconeer.exptool.a121._core.entities import (
    INT_16_COMPLEX,
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import Recorder

from .utils import PathOrH5File, h5_file_factory


T = TypeVar("T")


def get_h5py_str_dtype() -> Any:
    return h5py.special_dtype(vlen=str)


_H5PY_STR_DTYPE = get_h5py_str_dtype()


def get_timestamp() -> str:
    return datetime.datetime.now().isoformat(timespec="seconds")


def get_uuid() -> str:
    return str(uuid4())


class H5Recorder(Recorder):
    """Recorder writing directly to an HDF5 file

    :param path_or_file:
        If a path-like object, by default (depending on ``mode`` below) an HDF5 file is created at
        that path. When stopping, the file is closed.

        If an ``h5py.File``, that file is used as-is. In this case, the file is not closed when
        stopping.
    :param mode:
        The file mode to use if a path-like object was given for ``path_or_file``. Default value is
        'x', meaning that we open for exclusive creation, failing if the file already exists.
    :param _chunk_size:
        If given, data will be written to file every ``_chunk_size`` samples.

        If not given, data will be written at least every 512:th sample, or at least once per
        second, whichever comes first.

        Setting a small chunk size (e.g. 1) may degrade performance for high sample rates.

        Internal parameter, subject to change.
    """

    _AUTO_CHUNK_MAX_SIZE = 512
    _AUTO_CHUNK_MAX_TIME = 1.0

    path: Optional[os.PathLike]
    """The file path, if a path-like object was given for ``path_or_file``."""
    file: h5py.File
    """The ``h5py.File``."""
    owns_file: bool
    """Whether :class:`H5Recorder` opened and owns the file, i.e., if a path-like object was given
    for ``path_or_file``. If it does, the file is closed when stopping.
    """
    _num_frames: int
    _chunk_size: Optional[int]
    _chunk_buffer: list[list[dict[int, Result]]]
    _last_write_time: float

    def __init__(
        self,
        path_or_file: PathOrH5File,
        mode: str = "x",
        *,
        _chunk_size: Optional[int] = None,
        _lib_version: Optional[str] = None,
        _timestamp: Optional[str] = None,
        _uuid: Optional[str] = None,
    ) -> None:
        self.file, self.owns_file = h5_file_factory(path_or_file, h5_file_mode=mode)
        self.path = Path(self.file.filename) if self.owns_file else None
        self._chunk_size = _chunk_size
        self._chunk_buffer = []
        self._num_frames = 0
        self._last_write_time = 0.0

        if _lib_version is None:
            _lib_version = importlib_metadata.version("acconeer-exptool")

        if _timestamp is None:
            _timestamp = get_timestamp()

        if _uuid is None:
            _uuid = get_uuid()

        self.file.create_dataset(
            "uuid",
            data=_uuid,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "timestamp",
            data=_timestamp,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "lib_version",
            data=_lib_version,
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "generation",
            data="a121",
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

    def _start(
        self,
        *,
        client_info: ClientInfo,
        extended_metadata: list[dict[int, Metadata]],
        server_info: ServerInfo,
        session_config: SessionConfig,
        calibrations: Optional[dict[int, SensorCalibration]],
        calibrations_provided: Optional[dict[int, bool]],
    ) -> None:
        self.file.create_dataset(
            "client_info",
            data=client_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )
        self.file.create_dataset(
            "server_info",
            data=server_info.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

        session_group = self.file.create_group("session")

        session_group.create_dataset(
            "session_config",
            data=session_config.to_json(),
            dtype=_H5PY_STR_DTYPE,
            track_times=False,
        )

        for i, metadata_group_dict in enumerate(extended_metadata):
            group_group = session_group.create_group(f"group_{i}")

            for entry_id, (sensor_id, metadata) in enumerate(metadata_group_dict.items()):
                entry_group = group_group.create_group(f"entry_{entry_id}")
                entry_group.create_dataset("sensor_id", data=sensor_id, track_times=False)

                entry_group.create_dataset(
                    "metadata",
                    data=metadata.to_json(),
                    dtype=_H5PY_STR_DTYPE,
                    track_times=False,
                )

                result_group = entry_group.create_group("result")
                self._create_result_datasets(result_group, metadata)

        if (calibrations is None) != (calibrations_provided is None):
            raise ValueError(
                "'calibrations_provided' must be provided if 'calibrations' is provided"
            )

        if calibrations is not None and calibrations_provided is not None:
            calibrations_group = session_group.create_group("calibrations")
            for sensor_id, calibration in calibrations.items():
                sensor_calibration_group = calibrations_group.create_group(f"sensor_{sensor_id}")

                calibration.to_h5(sensor_calibration_group)

                sensor_calibration_group.create_dataset(
                    "provided", data=calibrations_provided[sensor_id], track_times=False
                )

        self._last_write_time = time()

    def _write_chunk_buffer_to_file(self, start_idx: int) -> int:
        """Saves the contents of ``self.chunk_buffer`` to file.

        :returns: the number of extended results saved.
        """

        if len(self._chunk_buffer) == 0:
            return 0

        for group_idx, entry_idx, results in utils.iterate_extended_structure_as_entry_list(
            utils.transpose_extended_structures(self._chunk_buffer)
        ):
            self._write_results(
                g=self.file[f"session/group_{group_idx}/entry_{entry_idx}/result"],
                start_index=start_idx,
                results=results,
            )

        return len(self._chunk_buffer)

    def _sample(self, extended_result: list[dict[int, Result]]) -> None:
        self._chunk_buffer.append(extended_result)

        if self._chunk_size is None:
            reached_size_limit = len(self._chunk_buffer) >= self._AUTO_CHUNK_MAX_SIZE
            reached_time_limit = (time() - self._last_write_time) >= self._AUTO_CHUNK_MAX_TIME
            write = reached_size_limit or reached_time_limit
        else:
            write = len(self._chunk_buffer) == self._chunk_size

        if write:
            self._num_frames += self._write_chunk_buffer_to_file(start_idx=self._num_frames)
            self._chunk_buffer = []
            self._last_write_time = time()

    def _stop(self) -> Any:
        self._num_frames += self._write_chunk_buffer_to_file(start_idx=self._num_frames)
        self._chunk_buffer = []

        if self.owns_file:
            self.file.close()

    @staticmethod
    def _create_result_datasets(g: h5py.Group, metadata: Metadata) -> None:
        g.create_dataset(
            "data_saturated",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "frame_delayed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "calibration_needed",
            shape=(0,),
            maxshape=(None,),
            dtype=bool,
            track_times=False,
            compression="gzip",
        )
        g.create_dataset(
            "temperature",
            shape=(0,),
            maxshape=(None,),
            dtype=int,
            track_times=False,
            compression="gzip",
        )

        g.create_dataset(
            "tick",
            shape=(0,),
            maxshape=(None,),
            dtype=np.dtype("int64"),
            track_times=False,
            compression="gzip",
        )

        g.create_dataset(
            "frame",
            shape=(0, *metadata.frame_shape),
            maxshape=(None, *metadata.frame_shape),
            dtype=INT_16_COMPLEX,
            track_times=False,
            compression="gzip",
        )

    @staticmethod
    def _write_results(g: h5py.Group, start_index: int, results: list[Result]) -> None:
        """Extends the Dataset to the appropriate (new) size with .resize,
        and then copies the data over
        """
        datasets_to_extend = [
            "data_saturated",
            "frame_delayed",
            "calibration_needed",
            "temperature",
            "tick",
            "frame",
        ]
        for dataset_name in datasets_to_extend:
            g[dataset_name].resize(size=start_index + len(results), axis=0)

        dataset_slice = slice(start_index, start_index + len(results))

        g["data_saturated"][dataset_slice] = [result.data_saturated for result in results]
        g["frame_delayed"][dataset_slice] = [result.frame_delayed for result in results]
        g["calibration_needed"][dataset_slice] = [result.calibration_needed for result in results]
        g["temperature"][dataset_slice] = [result.temperature for result in results]
        g["tick"][dataset_slice] = [result.tick for result in results]
        g["frame"][dataset_slice] = [result._frame for result in results]

    def require_algo_group(self, key: str) -> h5py.Group:
        """Creates/gets the ``algo`` group with a given ``key``.

        :raises: Exception if the key does not match the file content
        """

        group = self.file.require_group("algo")

        if "key" in group:
            existing_key = bytes(group["key"][()]).decode()
            if existing_key != key:
                raise Exception(f"Algo group key mismatch: got '{key}' but had '{existing_key}'")
        else:
            group.create_dataset(
                "key",
                data=key,
                dtype=_H5PY_STR_DTYPE,
                track_times=False,
            )

        return group



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import logging
import multiprocessing as mp
import platform
import queue
import signal
import socket
import traceback
from abc import ABCMeta, abstractmethod
from time import sleep, time

import serial

from acconeer.exptool._pyusb.pyusbcomm import PyUsbCdc
from acconeer.exptool.a111._clients.base import ClientError


try:
    from acconeer.exptool._winusbcdc.usb_cdc import ComPort
except ImportError:
    ComPort = None

log = logging.getLogger(__name__)


class LinkError(ClientError):
    pass


class BaseLink(metaclass=ABCMeta):
    DEFAULT_TIMEOUT = 2

    @abstractmethod
    def __init__(self):
        self._timeout = self.DEFAULT_TIMEOUT

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def recv(self, num_bytes):
        pass

    @abstractmethod
    def recv_until(self, bs):
        pass

    @abstractmethod
    def send(self, data):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, val):
        self._timeout = val
        self._update_timeout()

    def _update_timeout(self):
        pass


class SocketLink(BaseLink):
    _CHUNK_SIZE = 4096
    _PORT = 6110

    def __init__(self, host=None):
        super().__init__()
        self._host = host
        self._sock = None
        self._buf = None

    def _update_timeout(self):
        if self._sock is not None:
            self._sock.settimeout(self._timeout)

    def connect(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._update_timeout()

        try:
            self._sock.connect((self._host, self._PORT))
        except OSError as e:
            self._sock.close()
            self._sock = None
            raise LinkError("failed to connect") from e

        self._buf = bytearray()

    def recv(self, num_bytes):
        while len(self._buf) < num_bytes:
            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs):
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data

    def send(self, data):
        self._sock.sendall(data)

    def disconnect(self):
        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()
        self._sock = None
        self._buf = None


class BaseSerialLink(BaseLink):
    def __init__(self, baudrate=115200):
        super().__init__()
        self._timeout = self.DEFAULT_TIMEOUT
        self._baudrate = baudrate


class SerialLink(BaseSerialLink):
    def __init__(self, port=None, flowcontrol=False):
        super().__init__()
        self._port = port
        self._ser = None
        self._flowcontrol = flowcontrol

    def _update_timeout(self):
        if self._ser is not None:
            self._ser.timeout = self._timeout

    def send_break(self):
        self._ser.send_break()
        sleep(0.5)
        self._ser.flushInput()

    def connect(self):
        self._ser = serial.Serial()
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._update_timeout()
        self._ser.open()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

    def recv(self, num_bytes):
        data = bytearray(self._ser.read(num_bytes))

        if not len(data) == num_bytes:
            raise LinkError("recv timeout")

        return data

    def recv_until(self, bs):
        data = bytearray(self._ser.read_until(bs))

        if bs not in data:
            raise LinkError("recv timeout")

        return data

    def send(self, data):
        self._ser.write(data)

    def disconnect(self):
        self._ser.close()
        self._ser = None

    @property
    def baudrate(self):
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate):
        self._baudrate = new_baudrate

        if self._ser is not None and self._ser.is_open:
            self._ser.baudrate = new_baudrate


class ExploreSerialLink(SerialLink):
    _SERIAL_READ_PACKET_SIZE = 65536
    _SERIAL_PACKET_TIMEOUT = 0.01
    _SERIAL_WRITE_TIMEOUT = 1.0

    def __init__(self, port, flowcontrol=True):
        super().__init__(port, flowcontrol)
        self._buf = None

    def _update_timeout(self):
        pass

    def connect(self):
        self._ser = serial.Serial(
            timeout=self._SERIAL_PACKET_TIMEOUT,
            write_timeout=self._SERIAL_WRITE_TIMEOUT,
            exclusive=True,
        )
        self._ser.baudrate = self._baudrate
        self._ser.port = self._port
        self._ser.rtscts = self._flowcontrol
        self._ser.open()
        self._buf = bytearray()

        if platform.system().lower() == "windows":
            self._ser.set_buffer_size(rx_size=10**6, tx_size=10**6)

        self.send_break()

    def recv(self, num_bytes):
        t0 = time()
        while len(self._buf) < num_bytes:
            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs):
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._ser.read(self._SERIAL_READ_PACKET_SIZE))
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data


class USBLink(BaseLink):
    def __init__(self, vid=None, pid=None, serial=None):
        super().__init__()
        self._vid = vid
        self._pid = pid
        self._serial = serial

    def _update_timeout(self):
        pass

    def connect(self):
        # First try 'ComPort', will be set if platorm == windows
        port_type = ComPort
        if ComPort is None:
            # Fallback to  'PyUsbCdc', will be used if platform != windows
            port_type = PyUsbCdc

        if self._vid and self._pid:
            self._port = port_type(vid=self._vid, pid=self._pid, serial=self._serial, start=False)
        else:
            raise LinkError("Must have vid and pid for usb device")

        if not self._port.open():
            raise LinkError(f"Unable to connect to port (vid={self._vid}, pid={self._pid}")

        self._buf = bytearray()
        self.send_break()

    def send_break(self):
        self._port.send_break()
        sleep(0.5)
        self._port.reset_input_buffer()

    def recv(self, num_bytes):
        t0 = time()
        while len(self._buf) < num_bytes:
            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._port.read())
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs):
        t0 = time()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            try:
                r = bytearray(self._port.read())
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data

    def send(self, data):
        self._port.write(data)

    def disconnect(self):
        self._port.close()
        self._port = None


class SerialProcessLink(BaseSerialLink):
    def __init__(self, port=None):
        super().__init__()
        self._port = port
        self._process = None

    def connect(self):
        self._recv_queue = mp.Queue()
        self._send_queue = mp.Queue()
        self._flow_event = mp.Event()
        self._error_event = mp.Event()

        args = (
            self._port,
            self._baudrate,
            self._recv_queue,
            self._send_queue,
            self._flow_event,
            self._error_event,
        )

        self._process = mp.Process(
            target=serial_process_program,
            args=args,
            daemon=True,
        )

        self._process.start()

        flow_event_was_set = self._flow_event.wait(self._timeout)

        if flow_event_was_set:
            log.debug("connect - flow event was set")
        else:
            log.debug("connect - flow event was not set (timeout)")
            self.disconnect()
            raise LinkError("failed to connect, timeout")

        if self._error_event.is_set():
            log.debug("connect - error event was set")
            self.disconnect()
            raise LinkError("failed to connect, see traceback from serial process")

        log.debug("connect - successful")

        self._buf = bytearray()

    def recv(self, num_bytes):
        self.__empty_queue_into_buf()

        t0 = time()
        while len(self._buf) < num_bytes:
            self.__get_into_buf()

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs):
        self.__empty_queue_into_buf()

        n = len(bs)
        si = 0
        t0 = time()
        while True:
            buf_size = len(self._buf)
            if buf_size >= n:
                try:
                    i = self._buf.index(bs, si)
                except ValueError:
                    pass
                else:
                    break

                si = buf_size - n + 1

            if time() - t0 > self._timeout:
                raise LinkError("recv timeout")

            self.__get_into_buf()

        data = self._buf[: i + n]
        self._buf = self._buf[i + n :]
        return data

    def send(self, data):
        self._send_queue.put(data)

    def disconnect(self):
        if self._process.exitcode is None:
            self._flow_event.clear()
            self._process.join(1)

        if self._process.exitcode is None:
            log.info("forcing disconnect...")
            self._process.terminate()
            self._process.join(1)
            log.info("forced disconnect successful")

        if self._process.exitcode is None:
            raise LinkError("failed to disconnect")

    def __empty_queue_into_buf(self):
        while True:
            try:
                data = self._recv_queue.get_nowait()
            except queue.Empty:
                break

            self._buf.extend(data)

    def __get_into_buf(self):
        try:
            data = self._recv_queue.get(timeout=self._timeout)
        except queue.Empty:
            raise LinkError("recv timeout")
        except InterruptedError:
            return  # fixes interrupt issue on Windows

        self._buf.extend(data)

    @property
    def baudrate(self):
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate):
        self._baudrate = new_baudrate

        if self._process is not None and self._process.exitcode is None:
            log.debug("Changing baudrate to {}".format(new_baudrate))
            self._send_queue.put(("baudrate", new_baudrate))


def serial_process_program(port, baud, recv_q, send_q, flow_event, error_event):
    log.debug("serial communication process started")
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    try:
        _serial_process_program(port, baud, recv_q, send_q, flow_event, error_event)
    except Exception:
        error_event.set()
        flow_event.set()
        sleep(0.1)  # give the main process some time to print log messages
        print("Exception raised in serial process:\n")
        traceback.print_exc()
        print("\n\n")

    recv_q.close()
    send_q.close()


def _serial_process_program(port, baud, recv_q, send_q, flow_event, error_event):
    ser = serial.Serial(port=port, baudrate=baud, timeout=0, exclusive=True)
    flow_event.set()
    while flow_event.is_set():
        received = bytearray()
        while True:
            data = bytearray(ser.read(4096))
            if len(data) == 0:
                break
            received.extend(data)
        if len(received) > 0:
            recv_q.put(received)

        sent = False
        while True:
            try:
                x = send_q.get_nowait()
            except queue.Empty:
                break

            if isinstance(x, tuple):
                _, val = x  # assume its a baudrate change
                ser.baudrate = val
            else:
                ser.write(x)
                sent = True

        if not sent:
            sleep(0.0025)

    ser.close()



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from __future__ import annotations

import json
import typing as t

import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool.a121._core import complex_array_to_int16_complex

from ._processor import ProcessorResult


S = t.TypeVar("S")
T = t.TypeVar("T")

_ALL_RESULT_FIELDS: t.Final = (
    "frame",
    "distance_velocity_map",
    "amplitudes",
    "phases",
)


_INCOMPLETE_SERIALIZATION_MSG = (
    f"Group does not contains all fields {_ALL_RESULT_FIELDS}."
    + "Use 'allow_missing_fields' to replace missing fields with None"
)


def optional_apply(func: t.Callable[[S], T]) -> t.Callable[[t.Optional[S]], t.Optional[T]]:
    def wrapper(arg: t.Optional[S]) -> t.Optional[T]:
        if arg is None:
            return None
        else:
            return func(arg)

    return wrapper


_maybe_ndarray = optional_apply(np.array)


@optional_apply
def _frame_preprocessing(json_frame: npt.NDArray) -> npt.NDArray[np.complex_]:
    """
    A "json_frame" has the ".real" & ".imag" in a 2-lenght-list in the innermost dimension.
    """
    return json_frame[..., 0] + 1j * json_frame[..., 1]  # type: ignore[no-any-return]


class ProcessorResultJSONEncoder(json.JSONEncoder):
    def __init__(self, *, fields: t.Sequence[str] = _ALL_RESULT_FIELDS, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self.fields = fields

    def default(self, o: t.Any) -> t.Any:
        if isinstance(o, ProcessorResult):
            result = o
            full_dict = dict(
                frame=complex_array_to_int16_complex(result.frame).tolist(),
                distance_velocity_map=result.distance_velocity_map.tolist(),
                amplitudes=result.amplitudes.tolist(),
                phases=result.phases.tolist(),
            )
            return {k: v for k, v in full_dict.items() if k in self.fields}

        return super().default(o)


class ProcessorResultJSONDecoder(json.JSONDecoder):
    def __init__(self, *, allow_missing_fields: bool = False, **kwargs: t.Any) -> None:
        super().__init__(object_hook=self.object_hook, **kwargs)
        self.allow_missing_fields = allow_missing_fields

    def object_hook(self, obj: dict) -> ProcessorResult:
        frame = _frame_preprocessing(_maybe_ndarray(obj.get("frame")))
        dvm = _maybe_ndarray(obj.get("distance_velocity_map"))
        amplitudes = _maybe_ndarray(obj.get("amplitudes"))
        phases = _maybe_ndarray(obj.get("phases"))

        if not self.allow_missing_fields and (
            frame is None or dvm is None or amplitudes is None or phases is None
        ):
            raise ValueError(_INCOMPLETE_SERIALIZATION_MSG)

        return ProcessorResult(
            frame=frame,  # type: ignore[arg-type]
            distance_velocity_map=dvm,  # type: ignore[arg-type]
            amplitudes=amplitudes,  # type: ignore[arg-type]
            phases=phases,  # type: ignore[arg-type]
        )


class ProcessorResultJSONSerializer:
    def __init__(
        self, fields: t.Sequence = _ALL_RESULT_FIELDS, allow_missing_fields: bool = False
    ) -> None:
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, result: ProcessorResult) -> str:
        return json.dumps(result, cls=ProcessorResultJSONEncoder, fields=self.fields)

    def deserialize(self, result: str) -> ProcessorResult:
        return json.loads(  # type: ignore[no-any-return]
            result, cls=ProcessorResultJSONDecoder, allow_missing_fields=self.allow_missing_fields
        )


class ProcessorResultListH5Serializer:
    """
    Reads or writes a SparseIQ ProcessorResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: t.Sequence[str] = _ALL_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param destination: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on de-serialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during de-serialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    def serialize(self, results: t.List[ProcessorResult]) -> None:
        if "frame" in self.fields:
            self.group.create_dataset(
                "frame",
                dtype=complex,
                data=np.stack([res.frame for res in results]),
                track_times=False,
            )

        if "distance_velocity_map" in self.fields:
            self.group.create_dataset(
                "distance_velocity_map",
                dtype=float,
                data=np.stack([res.distance_velocity_map for res in results]),
                track_times=False,
            )

        if "amplitudes" in self.fields:
            self.group.create_dataset(
                "amplitudes",
                dtype=float,
                data=np.stack([res.amplitudes for res in results]),
                track_times=False,
            )

        if "phases" in self.fields:
            self.group.create_dataset(
                "phases",
                dtype=float,
                data=np.stack([res.phases for res in results]),
                track_times=False,
            )

    def _deserialize_at_index(self, index: int) -> ProcessorResult:
        frames = self.group.get("frame", None)
        dvms = self.group.get("distance_velocity_map", None)
        ampss = self.group.get("amplitudes", None)
        phasess = self.group.get("phases", None)

        if not self.allow_missing_fields and (
            frames is None or dvms is None or ampss is None or phasess is None
        ):
            raise ValueError(_INCOMPLETE_SERIALIZATION_MSG)

        frame = None if (frames is None) else frames[index]
        dvm = None if (dvms is None) else dvms[index]
        amps = None if (ampss is None) else ampss[index]
        phases = None if (phasess is None) else phasess[index]

        return ProcessorResult(
            frame=frame,  # type: ignore[arg-type]
            distance_velocity_map=dvm,  # type: ignore[arg-type]
            amplitudes=amps,  # type: ignore[arg-type]
            phases=phases,  # type: ignore[arg-type]
        )

    def deserialize(self, _: None) -> t.List[ProcessorResult]:
        optional_lens = set(
            optional_apply(len)(series)
            for series in [
                self.group.get("frame", None),
                self.group.get("distance_velocity_map", None),
                self.group.get("amplitudes", None),
                self.group.get("phases", None),
            ]
        )
        lens = {l for l in optional_lens if l is not None}

        (common_length,) = lens

        return [self._deserialize_at_index(i) for i in range(int(common_length))]



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import logging
from pathlib import Path
from typing import Callable, Generic, Optional, TypeVar

import h5py

import pyqtgraph as pg

from acconeer.exptool import a121
from acconeer.exptool.app.new import (
    AppModel,
    BackendPlugin,
    GeneralMessage,
    HandledException,
    Message,
    PlotPluginBase,
    PluginGeneration,
    PluginState,
    PluginStateMessage,
    ViewPluginBase,
    get_temp_h5_path,
    is_task,
)


log = logging.getLogger(__name__)


T = TypeVar("T")


class A121BackendPluginBase(Generic[T], BackendPlugin[T]):
    _live_client: Optional[a121.Client]
    _replaying_client: Optional[a121._ReplayingClient]
    _opened_record: Optional[a121.H5Record]
    _started: bool = False
    _recorder: Optional[a121.H5Recorder] = None
    _frame_count: int

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback, generation, key)
        self._live_client = None
        self._replaying_client = None
        self._opened_record = None
        self._frame_count = 0

    @is_task
    def load_from_file(self, *, path: Path) -> None:
        try:
            self._opened_record = a121.H5Record(h5py.File(path, mode="r"))
            self._replaying_client = a121._ReplayingClient(self._opened_record)
            self.load_from_record_setup(record=self._opened_record)
        except Exception as exc:
            self._opened_record = None
            self._replaying_client = None

            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            raise HandledException("Could not load from file") from exc

        self.start_session(with_recorder=False)

        self.send_status_message(f"<b>Replaying from {path.name}</b>")
        self.broadcast(sync=True)

    @abc.abstractmethod
    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        pass

    @is_task
    def load_from_cache(self) -> None:
        try:
            with self.h5_cache_file() as f:
                self._load_from_cache(f)
        except FileNotFoundError:
            pass

        self._sync_sensor_ids()
        self.broadcast(sync=True)

    @abc.abstractmethod
    def _load_from_cache(self, file: h5py.File) -> None:
        pass

    @abc.abstractmethod
    def get_next(self) -> None:
        pass

    def idle(self) -> bool:
        if self._started:
            if self.client is None:
                raise RuntimeError("Client is not attached. Can not 'get_next'.")

            try:
                self.get_next()
            except a121._StopReplay:
                self.stop_session()
                return True
            except Exception as exc:
                try:
                    self.stop_session()
                except Exception:
                    pass

                raise HandledException("Failed to get_next") from exc

            return True
        else:
            return False

    @property
    def client(self) -> Optional[a121.ClientBase]:
        if self._replaying_client is not None:
            return self._replaying_client

        return self._live_client

    @abc.abstractmethod
    def _sync_sensor_ids(self) -> None:
        pass

    def attach_client(self, *, client: a121.Client) -> None:
        self._live_client = client
        self._sync_sensor_ids()
        self.broadcast(True)

    def detach_client(self) -> None:
        self._live_client = None

    @abc.abstractmethod
    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        pass

    @abc.abstractmethod
    def end_session(self) -> None:
        pass

    @is_task
    def start_session(self, *, with_recorder: bool = True) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError("Client is not attached. Can not 'start'.")

        if not self.client.connected:
            raise RuntimeError("Client is not connected. Can not 'start'.")

        self.callback(GeneralMessage(name="saveable_file", data=None))
        if with_recorder:
            self._recorder = a121.H5Recorder(get_temp_h5_path())
        else:
            self._recorder = None

        try:
            self._start_session(self._recorder)
        except Exception as exc:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            raise HandledException("Could not start") from exc

        self._started = True

        self.broadcast()

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

    @is_task
    def stop_session(self) -> None:
        if not self._started:
            raise RuntimeError

        try:
            self.end_session()
        except Exception as exc:
            raise HandledException("Failure when stopping session") from exc
        finally:
            if self._recorder is not None:
                assert self._recorder.path is not None
                path = Path(self._recorder.path)
                self.callback(GeneralMessage(name="saveable_file", data=path))
                self._recorder = None

            self._started = False
            self._frame_count = 0
            self.broadcast()
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))
            self.callback(GeneralMessage(name="rate_stats", data=None))
            self.callback(GeneralMessage(name="frame_count", data=None))

        if self._opened_record:
            self._opened_record.close()
            self._opened_record = None
            self._replaying_client = None
            self._sync_sensor_ids()

    def teardown(self) -> None:
        try:
            with self.h5_cache_file(write=True) as file:
                self.save_to_cache(file)
        except Exception:
            log.warning("Could not write to cache")

        self.detach_client()

    @abc.abstractmethod
    def save_to_cache(self, file: h5py.File) -> None:
        pass


class A121ViewPluginBase(ViewPluginBase):
    pass


class A121PlotPluginBase(PlotPluginBase):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)
        self._is_setup = False
        self._plot_job: Optional[GeneralMessage] = None

    def handle_message(self, message: GeneralMessage) -> None:
        if message.name == "setup":
            self.plot_layout.clear()
            self.setup_from_message(message)
            self._is_setup = True
        elif message.name == "plot":
            self._plot_job = message
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.update_from_message(self._plot_job)
        finally:
            self._plot_job = None

    @abc.abstractmethod
    def setup_from_message(self, message: GeneralMessage) -> None:
        pass

    @abc.abstractmethod
    def update_from_message(self, message: GeneralMessage) -> None:
        pass



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

from enum import Enum, auto
from typing import Callable, Type

import numpy as np

from PySide6.QtWidgets import QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorPluginPreset,
    ProcessorViewPluginBase,
)
from acconeer.exptool.a121.algo._utils import get_distances_m
from acconeer.exptool.a121.algo.phase_tracking import (
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    pidgets,
)


class PluginPresetId(Enum):
    DEFAULT = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):

    PLUGIN_PRESETS = {
        PluginPresetId.DEFAULT.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_sensor_config()),
            processor_config=BackendPlugin.get_processor_config_cls()(),
        ),
    }

    @classmethod
    def get_processor_cls(cls) -> Type[Processor]:
        return Processor

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "threshold": pidgets.FloatParameterWidgetFactory(
                name_label_text="Threshold",
                decimals=1,
                limits=(0.0, 10000),
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig


class PlotPlugin(ProcessorPlotPluginBase[ProcessorResult]):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:
        (self.distances_m, _) = get_distances_m(sensor_config, metadata)

        pens = [et.utils.pg_pen_cycler(i) for i in range(3)]
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kws = [dict(pen=pen, **symbol_kw) for pen in pens]

        # sweep and threshold plot
        self.sweep_plot = self.plot_layout.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())
        self.sweeps_curve = [self.sweep_plot.plot(**feat_kw) for feat_kw in feat_kws]
        self.sweep_vertical_line = pg.InfiniteLine(pen=pens[2])
        self.sweep_plot.addItem(self.sweep_vertical_line)
        self.sweep_smooth_max = et.utils.SmoothMax()
        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweeps_curve[0], "Sweep")
        sweep_plot_legend.addItem(self.sweeps_curve[1], "Threshold")

        # argument plot
        argument_plot = self.plot_layout.addPlot(row=1, col=0)
        argument_plot.setMenuEnabled(False)
        argument_plot.showGrid(x=True, y=True)
        argument_plot.addLegend()
        argument_plot.setLabel("bottom", "Distance (m)")
        argument_plot.setLabel("left", "Phase")
        argument_plot.setYRange(-np.pi, np.pi)
        argument_plot.getAxis("left").setTicks(et.utils.pg_phase_ticks)
        argument_plot.setYRange(-np.pi, np.pi)
        argument_plot.addItem(pg.ScatterPlotItem())
        self.argument_curve = argument_plot.plot(
            **dict(pen=None, symbol="o", symbolSize=5, symbolPen="k")
        )
        self.argument_vertical_line = pg.InfiniteLine(pen=pens[2])
        argument_plot.addItem(self.argument_vertical_line)

        # history plot
        self.history_plot = self.plot_layout.addPlot(row=2, col=0)
        self.history_plot.setMenuEnabled(False)
        self.history_plot.showGrid(x=True, y=True)
        self.history_plot.addLegend()
        self.history_plot.setLabel("left", "Distance (mm)")
        self.history_plot.setLabel("bottom", "Time (s)")
        self.history_plot.addItem(pg.PlotDataItem())
        self.history_curve = self.history_plot.plot(**feat_kws[0])

        self.sweep_smooth_max = et.utils.SmoothMax()
        self.distance_hist_smooth_lim = et.utils.SmoothLimits(tau_decay=0.2, tau_grow=0.1)

    def update(self, processor_result: ProcessorResult) -> None:
        assert processor_result is not None
        assert processor_result.threshold is not None

        sweep = processor_result.lp_abs_sweep
        threshold = processor_result.threshold * np.ones(sweep.size)
        angle_sweep = processor_result.angle_sweep
        peak_loc = processor_result.peak_loc_m
        history = processor_result.distance_history
        rel_time_stamps = processor_result.rel_time_stamps

        # update sweep plot
        self.sweeps_curve[0].setData(self.distances_m, sweep)
        self.sweeps_curve[1].setData(self.distances_m, threshold)
        max_val_in_sweep_plot = max(np.max(sweep), np.max(threshold))
        self.sweep_plot.setYRange(0, self.sweep_smooth_max.update(max_val_in_sweep_plot))

        # update argument plot
        self.argument_curve.setData(self.distances_m, angle_sweep)

        if peak_loc is not None:
            # update vertical lines
            self.sweep_vertical_line.setValue(peak_loc)
            self.argument_vertical_line.setValue(peak_loc)
            self.sweep_vertical_line.show()
            self.argument_vertical_line.show()
        else:
            self.sweep_vertical_line.hide()
            self.argument_vertical_line.hide()

        if history.shape[0] != 0:
            # update history plot
            self.history_curve.setData(rel_time_stamps, history)
            lims = self.distance_hist_smooth_lim.update(history)
            self.history_plot.setYRange(lims[0], lims[1])
        self.history_plot.setXRange(-Processor.TIME_HORIZON_S, 0)


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPlugin:
        return ViewPlugin(app_model=app_model, view_widget=view_widget)

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPlugin:
        return PlotPlugin(app_model=app_model, plot_layout=plot_layout)


PHASE_TRACKING_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="phase_tracking",
    title="Phase tracking",
    description="Track target with micrometer precision.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from enum import Enum, unique
from typing import Optional, Type, TypeVar


T = TypeVar("T", bound=Enum)


def find_by_value(enum: Type[T], value: object) -> Optional[T]:
    for member in enum:
        if member.value == value:
            return member
    return None


def find_by_name(enum: Type[T], name: object) -> Optional[T]:
    for member in enum:
        if member.name == name:
            return member
    return None


def find_by_lowercase_name(enum: Type[T], lowercase_name: object) -> Optional[T]:
    for member in enum:
        if member.name.lower() == lowercase_name:
            return member
    return None


@unique
class Profile(Enum):
    """Profile"""

    PROFILE_1 = 1
    PROFILE_2 = 2
    PROFILE_3 = 3
    PROFILE_4 = 4
    PROFILE_5 = 5

    @classmethod
    def _missing_(cls, value: object) -> Optional[Profile]:
        return find_by_value(cls, value) or find_by_name(cls, value)


@unique
class PRF(Enum):
    """Pulse Repetition Frequency (PRF)"""

    PRF_19_5_MHz = 19500000
    PRF_15_6_MHz = 15600000
    PRF_13_0_MHz = 13000000
    PRF_8_7_MHz = 8700000
    PRF_6_5_MHz = 6500000
    PRF_5_2_MHz = 5200000

    @property
    def frequency(self) -> int:
        return self.value

    @classmethod
    def _missing_(cls, value: object) -> Optional[PRF]:
        return find_by_value(cls, value) or find_by_name(cls, value)


@unique
class IdleState(Enum):
    """Idle state"""

    DEEP_SLEEP = 0
    SLEEP = 1
    READY = 2

    @classmethod
    def _missing_(cls, value: object) -> Optional[IdleState]:
        if find_by_value(cls, value) is not None:
            return find_by_value(cls, value)

        if find_by_name(cls, value) is not None:
            return find_by_name(cls, value)

        if find_by_lowercase_name(cls, value) is not None:
            return find_by_lowercase_name(cls, value)

        return None

    def is_deeper_than(self, other: IdleState) -> bool:
        return self.value < other.value



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._configs import (
    get_close_and_far_processor_config,
    get_close_and_far_sensor_config,
    get_close_processor_config,
    get_close_sensor_config,
)
from ._processor import MeasurementType, Processor, ProcessorConfig, ProcessorResult



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from __future__ import annotations

import logging
from typing import Any, Optional

from acconeer.exptool.app.new.app_model import AppModel, AppModelListener


log = logging.getLogger(__name__)


class UiPluginBase(AppModelListener):
    """Extends AppModelListener by also stopping listening
    when plugins are unloaded (sig_load_plugin emits None).
    """

    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model)

        self.__app_model = app_model
        self.__app_model.sig_load_plugin.connect(self._on_load_plugin)
        log.debug(f"{self!s} is now listening to AppModel.")

    def stop_listening(self) -> None:
        super().stop_listening()
        self.__app_model.sig_load_plugin.disconnect(self._on_load_plugin)
        log.debug(f"{self!s} has stopped listening to AppModel.")

    def _on_load_plugin(self, plugin_spec: Optional[Any]) -> None:
        if plugin_spec is None:
            self.stop_listening()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import copy
import datetime
import json
import time
import warnings
from pathlib import Path
from typing import Optional, Union

import attr
import h5py
import numpy as np

import acconeer.exptool as et
from acconeer.exptool._structs import configbase
from acconeer.exptool.a111 import _configs, _modes


@attr.s
class Record:
    # Sensor session related (required):
    mode = attr.ib(type=_modes.Mode)  # save as str (Mode.name), restore with get_mode
    sensor_config_dump = attr.ib(type=str)  # SensorConfig._dumps
    session_info = attr.ib(type=dict)  # save/restore with json.dumps/loads
    data = attr.ib(default=None)  # [np.array], saved as np.array, restore as is
    data_info = attr.ib(type=list, factory=list)  # [[{...}]], save/restore with json.dumps/loads

    # Processing related (optional):
    module_key = attr.ib(type=Optional[str], default=None)
    processing_config_dump = attr.ib(type=Optional[str], default=None)  # ProcessingConfig._dumps

    # Other (optional):
    rss_version = attr.ib(type=Optional[str], default=None)
    lib_version = attr.ib(type=Optional[str], default=None)
    timestamp = attr.ib(type=Optional[str], default=None)
    sample_times = attr.ib(default=None)
    note = attr.ib(type=Optional[str], default=None)  # not to be used internally

    # Legacy (optional):
    legacy_processing_config_dump = attr.ib(type=Optional[str], default=None)

    def __attrs_post_init__(self):
        self._iter_index = None

    def __iter__(self):
        self._iter_index = 0
        return self

    def __next__(self):
        try:
            current_data_info = self.data_info[self._iter_index]
            current_data = self.data[self._iter_index]
        except (IndexError, TypeError):
            raise StopIteration

        self._iter_index += 1
        return current_data_info, current_data

    @property
    def sensor_config(self):
        return _configs.load(self.sensor_config_dump, self.mode)


class Recorder:
    def __init__(self, **kwargs):
        sensor_config = kwargs.pop("sensor_config")
        session_info = kwargs.pop("session_info")
        module_key = kwargs.pop("module_key", None)
        processing_config = kwargs.pop("processing_config", None)
        rss_version = kwargs.pop("rss_version", None)

        mode = kwargs.pop("mode", sensor_config.mode)

        self.max_len = kwargs.pop("max_len", None)

        if kwargs:
            key = next(iter(kwargs.keys()))
            msg = "Recorder got an unexpected keyword argument '{}'".format(key)
            raise TypeError(msg)

        if not isinstance(sensor_config, configbase.SensorConfig):
            raise TypeError("Unexpected sensor config type")

        if isinstance(processing_config, configbase.ProcessingConfig):
            processing_config_dump = processing_config._dumps()
        elif processing_config is None:
            processing_config_dump = None
        else:
            raise TypeError("Unexpected processing config type")

        self.record = Record(
            mode=mode,
            sensor_config_dump=sensor_config._dumps(),
            session_info=copy.deepcopy(session_info),
            module_key=module_key,
            processing_config_dump=processing_config_dump,
            rss_version=rss_version,
            lib_version=et.__version__,
            timestamp=datetime.datetime.now().isoformat(timespec="seconds"),
        )

        self.record.data = []
        self.record.sample_times = []

    def sample(self, data_info: list, data: np.ndarray):
        expected_num_dims = 3 if self.record.mode == _modes.Mode.SPARSE else 2
        if data.ndim != expected_num_dims:  # then assume data is squeezed
            # unsqueeze (add back sensor dim)
            data = data[None, ...]
            data_info = [data_info]

        self.record.data.append(data.copy())
        self.record.data_info.append(copy.deepcopy(data_info))

        self.record.sample_times.append(time.time())

        if self.max_len is not None and len(self.record.data) > self.max_len:
            self.record.data.pop(0)
            self.record.data_info.pop(0)
            self.record.sample_times.pop(0)

    def close(self):
        self.record.data = np.array(self.record.data)
        self.record.sample_times = np.array(self.record.sample_times)
        return self.record


def save(filename: Union[str, Path], record: Record):
    filename = str(filename)

    if filename.lower().endswith(".h5"):
        return save_h5(filename, record)
    elif filename.lower().endswith(".npz"):
        return save_npz(filename, record)
    elif filename.lower().endswith(".npy"):
        raise ValueError("Unknown file format '.npy', perhaps you meant '.npz'?")
    else:
        raise ValueError("Unknown file format")


def pack(record: Record) -> dict:
    packed = attr.asdict(record, filter=lambda attr, v: attr.type in (str, Optional[str]))
    packed["mode"] = record.mode.name.lower()
    packed["session_info"] = json.dumps(record.session_info)
    packed["data_info"] = json.dumps(record.data_info)

    data = np.array(record.data)
    if np.isrealobj(data):
        data_u16 = data.astype("u2")
        if np.all(data == data_u16):
            data = data_u16

    packed["data"] = data

    if record.sample_times is not None:
        packed["sample_times"] = np.array(record.sample_times)

    packed = {k: v for k, v in packed.items() if v is not None}

    return packed


def save_npz(filename: Union[str, Path], record: Record):
    filename = str(filename)

    if not filename.lower().endswith(".npz"):
        filename = filename + ".npz"

    packed = pack(record)
    np.savez_compressed(filename, **packed)


def save_h5(filename: Union[str, Path], record: Record):
    filename = str(filename)

    if not filename.lower().endswith(".h5"):
        filename = filename + ".h5"

    packed = pack(record)

    with h5py.File(filename, "w") as f:
        for k, v in packed.items():
            if isinstance(v, str):
                dtype = h5py.special_dtype(vlen=str)
                compression = None
            elif isinstance(v, np.ndarray):
                dtype = v.dtype
                compression = "gzip"
            else:
                raise TypeError

            f.create_dataset(k, data=v, dtype=dtype, compression=compression)


def load(filename: Union[str, Path]) -> Record:
    filename = str(filename)

    if filename.lower().endswith(".h5"):
        return load_h5(filename)
    elif filename.lower().endswith(".npz"):
        return load_npz(filename)
    else:
        raise ValueError("Unknown file format")


def unpack(packed: dict) -> Record:
    kwargs = {}

    data = packed["data"]
    if np.isrealobj(data):
        data = data.astype("float")

    kwargs["data"] = data

    for a in attr.fields(Record):
        k = a.name
        if a.type == str:
            kwargs[k] = packed[k]
        elif a.type == Optional[str]:
            kwargs[k] = packed.get(k, None)

    try:
        mode = _modes.get_mode(packed["mode"])
    except ValueError:
        mode = None
        warnings.warn("unknown mode encountered while unpacking record")

    kwargs["mode"] = mode

    kwargs["session_info"] = json.loads(packed["session_info"])
    kwargs["data_info"] = json.loads(packed["data_info"])

    kwargs["sample_times"] = packed.get("sample_times", None)

    assert len(kwargs["data"]) == len(kwargs["data_info"])

    return Record(**kwargs)


def load_npz(filename: Union[str, Path]) -> Record:
    filename = str(filename)

    packed = {}
    with np.load(filename, allow_pickle=False) as f:
        for k, v in f.items():
            if v.dtype.type is np.unicode_:
                v = str(v)

            packed[k] = v

    return unpack(packed)


def load_h5(filename: Union[str, Path]) -> Record:
    filename = str(filename)

    with h5py.File(filename, "r") as f:
        if "generation" in f:
            raise Exception(
                f"The file '{filename}' is not an A111 record, try a121.load_record instead"
            )
        packed = {k: v[()] for k, v in f.items()}

    for k, v in packed.items():
        if isinstance(v, bytes):
            packed[k] = v.decode()

    return unpack(packed)


if __name__ == "__main__":
    import argparse
    import os
    import sys

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    sp = subparsers.add_parser("resave")
    sp.add_argument("source")
    sp.add_argument("dest")
    sp.add_argument("-f", "--force", action="store_true")

    args = parser.parse_args()

    # assume resave as it's currently the only option

    if not args.force and os.path.exists(args.dest):
        sys.stderr.write("error: destination file already exists (try using -f)\n")
        sys.exit(1)

    record = load(args.source)
    save(args.dest, record)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import enum
import itertools
import json
import re
from functools import wraps
from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import attrs
import numpy as np
import packaging.version


S = TypeVar("S")
T = TypeVar("T")
U = TypeVar("U")

KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")

Number = Union[int, float]


class ProxyProperty(Generic[T]):
    """
    A descriptor that reduces boilerplate for accessing a certain property in a component.
    If the proxied property has a doc-string, it will be inherited by this descriptor.

    E.g. This can be used to implement "first-element-shortcuts"


        class Component:
            _attribute: int

            ...

            @property
            def attribute(self) -> int:
                return self._attribute

            @attribute.setter
            def attribute(self, new_value: int):
                self._attribute = new_value

        class Container:
            proxy_attribute = ProxyProperty[Component, int](
                lambda container: container.get_first_component(), Component.attribute
            )

            def get_first_component(self):
                ...

    It's recommended to subclass ProxyProperty to fit each Component-Container combination.
    """

    def __init__(
        self,
        accessor: Callable,
        prop: property,
    ):
        """
        :param accessor: The function to call to recieve an object that has the property "prop".
        :param prop: The property object which to proxy upon.
        """
        self._accessor = accessor
        if not isinstance(prop, property):
            raise TypeError(f'Passed "prop" needs to be a property. Actual type: {type(prop)}')
        self._property = prop

        self.__doc__ = prop.__doc__

    @overload
    def __get__(self, obj: None, objtype: Optional[Type] = ...) -> ProxyProperty[T]:
        ...

    @overload
    def __get__(self, obj: Any, objtype: Optional[Type] = ...) -> T:
        ...

    def __get__(
        self,
        obj: Optional[Any],
        objtype: Optional[Type] = None,
    ) -> Union[T, ProxyProperty[T]]:
        if obj is not None:
            proxee = self._accessor(obj)
            # The reason this needs to be ignored boils down to lack of support for
            # generic descriptors (property is a descriptor) in mypy.
            return self._property.__get__(proxee)  # type: ignore[no-any-return]

        return self

    def __set__(self, obj: Any, value: T) -> None:
        proxee = self._accessor(obj)
        self._property.__set__(proxee, value)


def unextend(structure: list[dict[int, T]]) -> T:
    """'Unexpands' a structure and returns the single element in the structure."""
    try:
        (group,) = structure
        (entry,) = group.values()
        return entry
    except Exception as e:
        raise ValueError(f"Could not unextend the structure {structure}") from e


def convert_value(value: Number, *, factory: Callable[[Number], T]) -> T:
    try:
        # May raise ValueError if e.g. "value" is a non-int string
        converted_value = factory(value)

        if converted_value != value:
            # E.g. int("3") != "3", int(3.5) != 3.5. is catched here.
            raise ValueError

        return converted_value
    except ValueError:
        raise TypeError(f"{value} cannot be converted with {factory}")


def _check_bounds(
    value: Number,
    lower_bound: Optional[Number] = None,
    upper_bound: Optional[Number] = None,
    inclusive: bool = True,
) -> None:
    """Raises a ValueError if:
    * ``value`` is not in [lower_bound, upper_bound], if inclusive = True
    * ``value`` is not in (lower_bound, upper_bound), if inclusive = False
    """
    exclusive = not inclusive

    boundaries = f"{lower_bound}, {upper_bound}"
    interval = f"({boundaries})" if exclusive else f"[{boundaries}]"
    error = ValueError(f"{value} needs to be in {interval}")

    if lower_bound is not None and (
        (exclusive and value <= lower_bound) or (inclusive and value < lower_bound)
    ):
        raise error

    if upper_bound is not None and (
        (exclusive and value >= upper_bound) or (inclusive and value > upper_bound)
    ):
        raise error


def convert_validate_int(
    value: Union[float, int], max_value: Optional[int] = None, min_value: Optional[int] = None
) -> int:
    """Converts an argument to an int.

    :param value: The argument to be converted and boundary checked
    :param max_value: Maximum value allowed
    :param min_value: Minimum value allowed

    :raises: TypeError if value is a string or a float with decimals
    :raises: ValueError if value does not agree with max_value and min_value
    """
    int_value = convert_value(value, factory=int)
    _check_bounds(int_value, lower_bound=min_value, upper_bound=max_value, inclusive=True)
    return int_value


def validate_float(
    value: float,
    max_value: Optional[float] = None,
    min_value: Optional[float] = None,
    inclusive: bool = True,
) -> float:
    """Converts an argument to a float.

    :param value: The argument to be converted and boundary checked
    :param max_value: Maximum value allowed
    :param min_value: Minimum value allowed
    :param inclusive:
        Whether the bounds ``max_value`` and ``min_value`` should be considered inclusive.
        E.g. value = 0.0, min_value = 0.0, inclusive = False raises a ValueError.

    :raises: TypeError if value cannot be converted to a float.
    :raises: ValueError if value does not agree with max_value and min_value
    """
    float_value = convert_value(value, factory=float)
    _check_bounds(float_value, lower_bound=min_value, upper_bound=max_value, inclusive=inclusive)
    return float_value


def is_multiple_of(multiplier: int, product: int) -> bool:
    """Returns True if `product` is a multiple of `multiplier`.
    I.e. checks if `multiplicand` is an integer in the equation

    `multiplicand` * `multiplier` = `product`
    """
    return product >= multiplier and product % multiplier == 0


def is_divisor_of(divisor: int, dividend: int) -> bool:
    """Returns True if `dividend` is divided by `divisor` with no remainder
    I.e. checks that `quotient` is an integer in the equation

    `dividend` / `divisor` = `quotient`
    """
    return is_multiple_of(dividend, divisor)


def map_over_extended_structure(
    func: Callable[[ValueT], T], structure: list[dict[KeyT, ValueT]]
) -> list[dict[KeyT, T]]:
    """Applies a function, `func`, to each element of the extended structure.

    Example:

        structure = [{1: "one"}, {2: "two"}]        # KeyT = int, ValueT = str
        func = str.encode                           # ValueT = str, T = bytes

        # Result
        result = [{1: b"one"}, {2: b"two"}]         # KeyT = int, T = bytes

    """
    return [{k: func(v) for k, v in d.items()} for d in structure]


def zip_extended_structures(
    structure_a: list[dict[int, S]],
    structure_b: list[dict[int, T]],
) -> list[dict[int, Tuple[S, T]]]:
    """Zip structures according to group id and sensor id.

    Typical use case is reordering extended structures before iterating.

    Example:

        structure_a = [{1: a_1, 2: a_2}, {1: a_3}]

        structure_b = [{1: b_1, 2: b_2}, {1: b_3}]

        result: [{1: (a_1, b_1), 2: (a_2, b_2)}, 1: {(a_3, b_3)}]

    """
    res = []
    try:
        for group_id, sensor_id, value in iterate_extended_structure(structure_a):
            res.append(
                (
                    group_id,
                    sensor_id,
                    (value, structure_b[group_id][sensor_id]),
                ),
            )
    except (KeyError, IndexError):
        raise ValueError("Structure of arguments are not the same.")

    return create_extended_structure(iter(res))


def zip3_extended_structures(
    structure_a: list[dict[int, S]],
    structure_b: list[dict[int, T]],
    structure_c: list[dict[int, U]],
) -> list[dict[int, Tuple[S, T, U]]]:
    """Zip structures according to group id and sensor id.

    Typical use case is reordering extended structures before iterating.

    Example:

        structure_a = [{1: a_1, 2: a_2}, {1: a_3}]

        structure_b = [{1: b_1, 2: b_2}, {1: b_3}]

        structure_c = [{1: c_1, 2: c_2}, {1: c_3}]

        result: [{1: (a_1, b_1, c_1), 2: (a_2, b_2, c_2)}, {1: (a_3, b_3, c_3)}]

    """
    res = []
    try:
        for group_id, sensor_id, value in iterate_extended_structure(structure_a):
            res.append(
                (
                    group_id,
                    sensor_id,
                    (value, structure_b[group_id][sensor_id], structure_c[group_id][sensor_id]),
                ),
            )
    except (KeyError, IndexError):
        raise ValueError("Structure of arguments are not the same.")

    return create_extended_structure(iter(res))


def iterate_extended_structure(
    structure: list[dict[int, ValueT]]
) -> Iterator[Tuple[int, int, ValueT]]:
    """Iterates over the elements of the extended structure.

    :returns: Iterator of (<group id>, <sensor id>, <element>)
    """

    for group_id, group in enumerate(structure):
        for sensor_id, elem in group.items():
            yield (group_id, sensor_id, elem)


def iterate_extended_structure_as_entry_list(
    structure: list[dict[int, ValueT]]
) -> Iterator[Tuple[int, int, ValueT]]:
    """Iterates over the elements of the extended structure.

    Note: this differs from ``iterate_extended_structure`` as it returns the
    ``entry_idx`` instead of the sensor id of a given entry.

    :returns: Iterator of (<group idx>, <entry idx>, <element>)
    """

    for group_idx, group in enumerate(structure):
        for entry_idx, elem in enumerate(group.values()):
            yield (group_idx, entry_idx, elem)


def extended_structure_entry_count(structure: list[dict[int, Any]]) -> int:
    """Traverses the extended structure and returns a count of all its entries

    :returns: number of entries
    """

    return sum(len(group) for group in structure)


def iterate_extended_structure_values(structure: list[dict[int, ValueT]]) -> Iterator[ValueT]:
    """Iterates like `iterate_extended_structure` but throws away group id and sensor id."""
    for _, _, value in iterate_extended_structure(structure):
        yield value


def extended_structure_shape(structure: list[dict[int, Any]]) -> list[set[int]]:
    return [set(group.keys()) for group in structure]


def transpose_extended_structures(
    structures: list[list[dict[int, ValueT]]]
) -> list[dict[int, list[ValueT]]]:
    """'Transposes' a list of extended structures to create an extended structure of lists"""

    if not structures:
        raise ValueError("'structures' cannot be empty")

    shapes = [extended_structure_shape(s) for s in structures]
    if not all(shape == shapes[0] for shape in shapes):
        raise ValueError("All extended structures needs to have the same structure.")

    product: list[dict[int, list[ValueT]]] = map_over_extended_structure(
        lambda _: list(), structures[0]
    )

    for group_idx, sensor_id, value in itertools.chain(
        *[iterate_extended_structure(es) for es in structures]
    ):
        product[group_idx][sensor_id].append(value)

    return product


def create_extended_structure(items: Iterator[Tuple[int, int, ValueT]]) -> list[dict[int, ValueT]]:
    structure: list[dict[int, ValueT]] = []
    current_group_index: Optional[int] = None
    current_group: Optional[dict[int, ValueT]] = None

    for group_index, sensor_id, value in items:
        if current_group_index is None:
            if group_index != 0:
                raise ValueError

            current_group_index = 0
            current_group = {}
            structure.append(current_group)
        elif group_index != current_group_index:
            if group_index != current_group_index + 1:
                raise ValueError

            current_group_index += 1
            current_group = {}
            structure.append(current_group)

        assert current_group is not None
        if sensor_id in current_group:
            raise ValueError

        current_group[sensor_id] = value

    return structure


class EntityJSONEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, enum.Enum):
            return obj.name

        return json.JSONEncoder.default(self, obj)


def parse_rss_version(rss_version: str) -> packaging.version.Version:
    """Takes an RSS version string and returns a corresponding Version

    The RSS version string is on a 'git describe'-like format:

        a121-vA.B.C<-rcD><-E-gF>

    where

        A: major, B: minor, C: micro,
        D: release candidate,
        E: additional commits since tag, F: commit SHA

    The concept of 'additional commits since tag' (E) doesn't have an
    equivalent in packaging.version.Version. Instead, when E is present,
    the smallest version part (D if present, otherwise C) is bumped and
    E is presented as a development prerelease.

    The commit SHA (F), if present, is translated to a 'local segment'.

    Examples:

        "a121-v1.2.3" ->
            Version("1.2.3")
        "a121-v1.2.3-rc4" ->
            Version("1.2.3rc4")
        "a121-v1.2.3-123-g0e03503be1" ->
            Version("1.2.4.dev123+g0e03503be1")
        "a121-v1.2.3-rc4-123-g0e03503be1" ->
            Version("1.2.3rc5.dev123+g0e03503be1")

    Read more: https://packaging.pypa.io/en/latest/version.html
    """

    pattern = (
        r"a121-v(?P<major>\d+)\.(?P<minor>\d+)\.(?P<micro>\d+)"
        r"(?:-(?P<pre_phase>rc)(?P<pre_number>\d+))?"
        r"(?:-(?P<dev_number>\d+)-(?P<dev_commit>g\w+))?"
        r"(?:-(dirty))?"
    )
    match = re.fullmatch(pattern, rss_version)
    if not match:
        raise ValueError("Not a valid RSS version")

    groups = match.groupdict()

    is_prerelease = groups["pre_number"] is not None
    is_devrelease = groups["dev_number"] is not None

    release_segment = ""
    pre_segment = ""
    dev_segment = ""

    if is_devrelease:
        dev_segment = f".dev{groups['dev_number']}+{groups['dev_commit']}"

        if is_prerelease:
            groups["pre_number"] = int(groups["pre_number"]) + 1
        else:
            groups["micro"] = int(groups["micro"]) + 1

    if is_prerelease:
        pre_segment = f"{groups['pre_phase']}{groups['pre_number']}"

    release_segment = f"{groups['major']}.{groups['minor']}.{groups['micro']}"

    version = release_segment + pre_segment + dev_segment
    return packaging.version.Version(version)


def unwrap_ticks(
    ticks: list[int], minimum_tick: Optional[int], limit: int = 2**32
) -> Tuple[list[int], Optional[int]]:
    """Unwraps a sequence of ticks belonging to a extended result

    The server tick (attached to every produced Result) wraps at 2^32. Thus, if the raw tick is
    used for evaluating the time between two results, it will be incorrect if a wrap has occurred
    between them. Therefore, it has to be accounted for by "unwrapping".

    Wrapping can occur between the results in an extended result, and that the results are not
    necessarily ordered by the tick. This means that we have to look at all the ticks produced in
    the extended result at the same time.

    For example:

    Let's say the wrap happens at limit = 100, and that we have an extended result of two elements
    with ticks 10 and 90. Since it's more likely that 10 has wrapped than not, we assume it's
    actually after 90, i.e., 110.

    Now, let's also consider that the previous maximum unwrapped tick was 195, which is now the
    'minimum tick'. From this, we know that the ticks have wrapped before and can account for that,
    resulting in the final unwrapped ticks of 310 and 290.
    """

    if len(ticks) == 0:
        return [], None

    if any(tick < 0 or tick >= limit for tick in ticks):
        raise ValueError("Tick value out of bounds")

    if (max(ticks) - min(ticks)) > limit // 2:
        ticks = [tick + limit if tick < limit // 2 else tick for tick in ticks]

    if minimum_tick is not None:
        num_wraps = max((minimum_tick - tick - 1) // limit + 1 for tick in ticks)
        ticks = [num_wraps * limit + tick for tick in ticks]

    return ticks, max(ticks)


def pretty_dict_line_strs(d: dict[str, Any], indent: int = 2, width: int = 24) -> list[str]:
    lines = []
    for k, v in d.items():
        if isinstance(v, enum.Enum):
            v = v.name

        lines.append(f"{'':<{indent}}{k + ' ':.<{width}} {v}")

    return lines


def indent_strs(strs: list[str], level: int) -> list[str]:
    return ["  " * level + s for s in strs]


attrs_ndarray_eq = attrs.cmp_using(eq=np.array_equal)
attrs_ndarray_isclose = attrs.cmp_using(eq=lambda a, b: bool(np.isclose(a, b).all()))


def no_dynamic_member_creation(cls: Type[T]) -> Type[T]:
    """
    Class annotation that prevents setting any attributes that does not exist
    after the object have been created (the __init__ function have executed).
    """

    def setattr_wrapper(func):  # type: ignore
        @wraps(func)
        def setattr(self, key, value):  # type: ignore
            if hasattr(self, "__frozen") and not hasattr(self, key):
                raise AttributeError(f'Invalid attribute "{key}" for {self!r}')
            else:
                func(self, key, value)

        return setattr

    def init_wrapper(func):  # type: ignore
        @wraps(func)
        def init(self, *args, **kwargs):  # type: ignore
            func(self, *args, **kwargs)
            self.__frozen = True

        return init

    cls.__setattr__ = setattr_wrapper(cls.__setattr__)  # type: ignore
    cls.__init__ = init_wrapper(cls.__init__)  # type: ignore
    return cls



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from PySide6.QtCore import QObject, QThread, QTimerEvent, Signal, Slot


if not TYPE_CHECKING:
    from acconeer.exptool.utils import get_serial_devices, get_usb_devices


class PortUpdater(QObject):
    sig_update = Signal(object, object)

    class Worker(QObject):
        sig_update = Signal(object, object)
        serial_devices: Any = None
        usb_devices: Any = None

        @Slot()  # type: ignore[misc]
        def start(self) -> None:
            self.timer_id = self.startTimer(500)

        @Slot()  # type: ignore[misc]
        def stop(self) -> None:
            self.killTimer(self.timer_id)

        def timerEvent(self, event: QTimerEvent) -> None:
            serial_devices = get_serial_devices()  # type: ignore[name-defined]
            usb_devices = get_usb_devices()  # type: ignore[name-defined]

            if self.serial_devices != serial_devices or self.usb_devices != usb_devices:
                self.serial_devices = serial_devices
                self.usb_devices = usb_devices
                self.sig_update.emit(serial_devices, usb_devices)

    def __init__(self, parent: QObject) -> None:
        super().__init__(parent)

        self.thread = QThread(self)
        self.worker = self.Worker()
        self.thread.started.connect(self.worker.start)
        self.thread.finished.connect(self.worker.stop)
        self.worker.sig_update.connect(self._on_update)
        self.worker.moveToThread(self.thread)
        self.signalling = False

    def start(self) -> None:
        self.thread.start()
        self.signalling = True

    def stop(self) -> None:
        self.thread.quit()
        self.thread.wait()
        self.signalling = False

    def pause(self) -> None:
        self.signalling = False

    def resume(self) -> None:
        self.signalling = True

    def _on_update(self, serial: Any, usb: Any) -> None:
        if self.signalling:
            self.sig_update.emit(serial, usb)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
from typing import Callable, Generic, Mapping, Optional, Type

import attrs
import h5py

from acconeer.exptool import a121
from acconeer.exptool.a121 import _core
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._base import (
    GenericProcessorBase,
    InputT,
    MetadataT,
    ProcessorConfigT,
    ResultT,
)
from acconeer.exptool.a121.algo._plugins._a121 import A121BackendPluginBase
from acconeer.exptool.app.new import (
    BackendLogger,
    GeneralMessage,
    Message,
    PluginGeneration,
    is_task,
)


CALIBRATION_NEEDED_MESSAGE = "Calibration needed - restart"
DATA_SATURATED_MESSAGE = "Data saturated - reduce gain"
FRAME_DELAYED_MESSAGE = "Frame delayed"


@attrs.mutable(kw_only=True)
class ProcessorPluginPreset(Generic[ProcessorConfigT]):
    session_config: a121.SessionConfig = attrs.field()
    processor_config: ProcessorConfigT = attrs.field()


@attrs.mutable(kw_only=True)
class ProcessorBackendPluginSharedState(Generic[ProcessorConfigT, MetadataT]):
    session_config: a121.SessionConfig = attrs.field()
    processor_config: ProcessorConfigT = attrs.field()
    replaying: bool = attrs.field(default=False)
    metadata: Optional[MetadataT] = attrs.field(default=None)

    @property
    def ready(self) -> bool:
        try:
            self.session_config.validate()
            self.processor_config.validate(self.session_config)
        except a121.ValidationError:
            return False
        else:
            return True


class GenericProcessorBackendPluginBase(
    Generic[InputT, ProcessorConfigT, ResultT, MetadataT],
    A121BackendPluginBase[ProcessorBackendPluginSharedState[ProcessorConfigT, MetadataT]],
):
    _processor_instance: Optional[
        GenericProcessorBase[InputT, ProcessorConfigT, ResultT, MetadataT]
    ]
    _recorder: Optional[a121.H5Recorder]
    _started: bool

    PLUGIN_PRESETS: Mapping[int, Callable[[], ProcessorPluginPreset]] = {}

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ):
        super().__init__(callback=callback, generation=generation, key=key)
        self._processor_instance = None
        self._recorder = None
        self._log = BackendLogger.getLogger(__name__)
        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.session_config = a121.SessionConfig.from_json(file["session_config"][()])
        self.shared_state.processor_config = self.get_processor_config_cls().from_json(
            file["processor_config"][()]
        )

    @is_task
    @abc.abstractmethod
    def restore_defaults(self) -> None:
        pass

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if (
                len(sensor_ids) > 0
                and self.shared_state.session_config.sensor_id not in sensor_ids
            ):
                self.shared_state.session_config.sensor_id = sensor_ids[0]

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(
            file, "session_config", self.shared_state.session_config.to_json()
        )
        _create_h5_string_dataset(
            file, "processor_config", self.shared_state.processor_config.to_json()
        )

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        self.shared_state.session_config = record.session_config
        algo_group = record.get_algo_group(self.key)  # noqa: F841
        self.shared_state.processor_config = self.get_processor_config_cls().from_json(
            algo_group["processor_config"][()]
        )

    @is_task
    def update_session_config(self, *, session_config: a121.SessionConfig) -> None:
        self.shared_state.session_config = session_config
        self.broadcast()

    @is_task
    def update_processor_config(self, *, processor_config: ProcessorConfigT) -> None:
        self.shared_state.processor_config = processor_config
        self.broadcast()

    def _validate_session_config(self, session_config: a121.SessionConfig) -> None:
        """Hook for validating session configs."""
        pass

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        session_config = self.shared_state.session_config
        self._validate_session_config(session_config)

        if recorder:
            algo_group = recorder.require_algo_group(self.key)  # noqa: F841

            # TODO: break out saving (?)
            _create_h5_string_dataset(
                algo_group, "processor_config", self.shared_state.processor_config.to_json()
            )
        assert self.client
        metadata = self.client.setup_session(session_config)
        the_first_sensor_config = next(
            _core.utils.iterate_extended_structure_values(session_config.groups)
        )
        self._processor_instance = self.get_processor_cls()(
            sensor_config=the_first_sensor_config,  # FIXME: a bit scuffed but wth
            metadata=metadata,  # type: ignore[arg-type]
            processor_config=self.shared_state.processor_config,
        )

        self.shared_state.metadata = metadata  # type: ignore[assignment]
        self.client.start_session(self._recorder)

        self.callback(
            GeneralMessage(
                name="setup",
                kwargs=dict(metadata=metadata, sensor_config=the_first_sensor_config),
                recipient="plot_plugin",
            )
        )

    def end_session(self) -> None:
        if self.client is None:
            raise RuntimeError("Client is not attached. Can not 'stop'.")
        self.client.stop_session()

    @classmethod
    @abc.abstractmethod
    def get_processor_cls(
        cls,
    ) -> Type[GenericProcessorBase[InputT, ProcessorConfigT, ResultT, MetadataT]]:
        pass

    @classmethod
    @abc.abstractmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfigT]:
        pass

    @classmethod
    @abc.abstractmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        pass

    @staticmethod
    def _format_warning(s: str) -> str:
        return f'<p style="color: #FD5200;"><b>Warning: {s}</b></p>'


class ProcessorBackendPluginBase(
    GenericProcessorBackendPluginBase[a121.Result, ProcessorConfigT, ResultT, a121.Metadata]
):
    def _validate_session_config(self, session_config: a121.SessionConfig) -> None:
        if session_config.extended:
            raise ValueError("Extended session configs are not supported.")

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = ProcessorBackendPluginSharedState(
            session_config=a121.SessionConfig(self.get_default_sensor_config()),
            processor_config=self.get_processor_config_cls()(),
        )

        self.broadcast(sync=True)

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        processor_preset = preset_config()
        self.shared_state.session_config = processor_preset.session_config
        self.shared_state.processor_config = processor_preset.processor_config
        self.broadcast(sync=True)

    def get_next(self) -> None:
        if self._processor_instance is None:
            raise RuntimeError("Processor is None. 'start' needs to be called before 'get_next'")

        assert self.client
        result = self.client.get_next()
        assert isinstance(result, a121.Result)  # TODO: fix

        self._frame_count += 1

        if result.data_saturated:
            self.send_status_message(self._format_warning(DATA_SATURATED_MESSAGE))

        if result.calibration_needed:
            self.send_status_message(self._format_warning(CALIBRATION_NEEDED_MESSAGE))

        if result.frame_delayed:
            self.send_status_message(self._format_warning(FRAME_DELAYED_MESSAGE))

        processor_result = self._processor_instance.process(result)
        self.callback(GeneralMessage(name="rate_stats", data=self.client._rate_stats))
        self.callback(GeneralMessage(name="frame_count", data=self._frame_count))
        self.callback(GeneralMessage(name="plot", data=processor_result, recipient="plot_plugin"))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Optional, Union

import attrs
from typing_extensions import Literal

from acconeer.exptool.app.new._enums import ConnectionState, PluginState


RecipientLiteral = Union[Literal["plot_plugin"], Literal["view_plugin"]]


@attrs.frozen(kw_only=True, slots=False)
class Message:  # Should not be instantiated
    pass


@attrs.frozen(kw_only=True, slots=False)
class ConnectionStateMessage(Message):
    state: ConnectionState = attrs.field()
    warning: Optional[str] = attrs.field(default=None)


@attrs.frozen(kw_only=True, slots=False)
class PluginStateMessage(Message):
    state: PluginState = attrs.field()


@attrs.frozen(kw_only=True, slots=False)
class BackendPluginStateMessage(Message):
    state: Any = attrs.field()


@attrs.frozen(kw_only=True, slots=False)
class StatusMessage(Message):
    status: Optional[str] = attrs.field()


@attrs.frozen(kw_only=True, slots=False)
class LogMessage(Message):
    module_name: str = attrs.field()
    log_level: str = attrs.field()
    log_string: str = attrs.field()


@attrs.frozen(kw_only=True, slots=False)
class GeneralMessage(Message):
    name: str = attrs.field()
    recipient: Optional[RecipientLiteral] = attrs.field(default=None)
    data: Optional[Any] = attrs.field(default=None)
    kwargs: Optional[dict[str, Any]] = attrs.field(default=None)
    exception: Optional[Exception] = attrs.field(default=None)
    traceback_format_exc: Optional[str] = attrs.field(default=None)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from collections import namedtuple

from PySide6 import QtCore, QtWidgets
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSpinBox,
    QToolButton,
    QVBoxLayout,
    QWidget,
)


class AdvancedSerialDialog(QDialog):
    def __init__(self, state, parent):
        super().__init__(parent)

        self.setMinimumWidth(350)
        self.setModal(True)
        self.setWindowTitle("Advanced serial settings")

        layout = QVBoxLayout()
        self.setLayout(layout)

        texts = [
            "Please note:",
            (
                "Overriding the baud rate disables automatic"
                " detection and negotiation of baud rate."
            ),
            "Only use on special hardware.",
        ]

        for text in texts:
            lbl = QLabel(text, self)
            lbl.setWordWrap(True)
            layout.addWidget(lbl)

        layout.addStretch(1)

        self.cb = QCheckBox("Override baud rate", self)
        self.cb.stateChanged.connect(self.cb_state_changed)
        layout.addWidget(self.cb)

        self.sb = QSpinBox(self)
        self.sb.setRange(1, int(3e6))
        layout.addWidget(self.sb)

        layout.addStretch(1)

        buttons_widget = QWidget(self)
        layout.addWidget(buttons_widget)
        hbox = QHBoxLayout()
        buttons_widget.setLayout(hbox)
        hbox.addStretch(1)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        hbox.addWidget(cancel_btn)
        save_btn = QPushButton("Save")
        save_btn.setDefault(True)
        save_btn.clicked.connect(self.accept)
        hbox.addWidget(save_btn)

        self.set_state(state)

    def cb_state_changed(self, state):
        self.sb.setEnabled(bool(state))

    def set_state(self, state):
        checked = state is not None
        self.cb.setChecked(checked)
        self.cb_state_changed(checked)
        self.sb.setValue(state if checked else 115200)

    def get_state(self):
        return self.sb.value() if self.cb.checkState() else None


class BiggerMessageBox(QtWidgets.QMessageBox):
    def resizeEvent(self, event):
        result = super().resizeEvent(event)
        self.setFixedWidth(500)
        return result


class Label(QLabel):
    def __init__(self, img, img_scale=0.7):
        super(Label, self).__init__()

        self.img_scale = img_scale
        self.pixmap = QPixmap(img)

        self.setMinimumSize(1, 1)
        self.setAlignment(QtCore.Qt.AlignVCenter | QtCore.Qt.AlignHCenter)
        self.setPixmap(self.pixmap)

    def resizeEvent(self, event):
        w = int(round(self.size().width() * self.img_scale))
        h = int(round(self.size().height() * self.img_scale))
        scaled_size = QtCore.QSize(w, h)

        scaled_pixmap = self.pixmap.scaled(
            scaled_size,
            QtCore.Qt.KeepAspectRatio,
            QtCore.Qt.SmoothTransformation,
        )

        self.setPixmap(scaled_pixmap)


class CollapsibleSection(QFrame):
    def __init__(self, header_text, init_collapsed=False, is_top=False):
        super().__init__()

        if not is_top:
            self.setObjectName("CollapsibleSection")
            self.setStyleSheet("#CollapsibleSection{border-top: 1px solid lightgrey;}")

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setLayout(self._layout)
        self._header_widget = QWidget()
        self.body_widget = QWidget()
        self._layout.addWidget(self._header_widget)
        self._layout.addWidget(self.body_widget)

        self.grid = QGridLayout(self.body_widget)
        self.grid.setContentsMargins(9, 0, 9, 9)
        self.grid.setColumnStretch(0, 1)
        self.grid.setColumnStretch(1, 1)

        self._header_widget_layout = QHBoxLayout(self._header_widget)
        self._header_widget_layout.setContentsMargins(7, 7, 7, 7)
        self._header_widget.setLayout(self._header_widget_layout)

        self._button = QToolButton()
        self._button.setText(header_text)
        self._button.setCheckable(True)
        self._button.setStyleSheet("QToolButton { border: none; }")
        self._button.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self._button.pressed.connect(self.button_event)
        self.button_event(override=init_collapsed)
        self._header_widget_layout.addWidget(self._button)
        self._header_widget_layout.addStretch()

    def button_event(self, override=None):
        if override is None:
            checked = not self._button.isChecked()
        else:
            checked = override
            self._button.setChecked(checked)

        if checked:  # collapsed
            self._button.setArrowType(QtCore.Qt.ArrowType.RightArrow)
            self.body_widget.hide()
        else:
            self._button.setArrowType(QtCore.Qt.ArrowType.DownArrow)
            self.body_widget.show()


class SensorSelection(QFrame):
    def __init__(self, multi_sensors=False, error_handler=None, callback=None):
        super().__init__()

        self.error_handler = error_handler
        self.callback = callback

        self.drawing = False
        self.selected = [1]
        self.sources = None
        self.module_multi_sensor_support = multi_sensors

        self.grid = QtWidgets.QGridLayout(self)
        self.grid.setContentsMargins(0, 0, 0, 0)

        self.checkboxes = []
        self.radio_buttons = []
        for i in range(4):
            s = i + 1

            cb = QCheckBox(str(s), self)
            cb.stateChanged.connect(lambda val, cb=cb: self.checkbox_event_handler(val, cb))
            self.checkboxes.append(cb)
            self.grid.addWidget(cb, 0, i)

            rb = QtWidgets.QRadioButton(str(s), self)
            rb.toggled.connect(lambda val, rb=rb: self.radio_button_event_handler(val, rb))
            self.radio_buttons.append(rb)
            self.grid.addWidget(rb, 1, i)

        self.draw()

    def checkbox_event_handler(self, state, cb):
        if self.drawing:
            return

        s = int(cb.text())

        if state:  # checked
            if s not in self.selected:
                self.selected.insert(0, s)
        else:
            if s in self.selected:
                self.selected.remove(s)

        self.draw()

    def radio_button_event_handler(self, selected, rb):
        if self.drawing or not selected:
            return

        s = int(rb.text())
        self.selected = [s]
        self.draw()

    def draw(self):
        self.drawing = True

        for i, (cb, rb) in enumerate(zip(self.checkboxes, self.radio_buttons)):
            s = i + 1

            if self.sources:
                enabled = s in self.sources
            else:
                enabled = True

            cb.setEnabled(enabled)
            rb.setEnabled(enabled)

            cb.setVisible(bool(self.module_multi_sensor_support))
            rb.setVisible(not bool(self.module_multi_sensor_support))

        for i, cb in enumerate(self.checkboxes):
            s = i + 1
            cb.setChecked(s in self.selected)
            rb.setChecked(False)

        if self.selected:
            self.radio_buttons[self.selected[0] - 1].setChecked(True)

        self.drawing = False

        if self.callback is not None:
            self.callback()

    def sanitize(self):
        if self.sources:
            self.selected = [s for s in self.selected if s in self.sources]

        if self.module_multi_sensor_support:
            if isinstance(self.module_multi_sensor_support, list):
                lim = self.module_multi_sensor_support[1]
                self.selected = self.selected[:lim]
        else:
            if self.selected:
                self.selected = [self.selected[0]]

    def get_sensors(self):
        return list(sorted(self.selected))

    def set_sensors(self, sensors):
        self.selected = sensors
        self.sanitize()
        self.draw()

    def set_multi_sensor_support(self, sources, module_multi_sensor_support):
        self.sources = sources
        self.module_multi_sensor_support = module_multi_sensor_support
        self.sanitize()
        self.draw()


class SessionInfoView(QWidget):
    Field = namedtuple("Field", ["label", "unit", "fmt_str", "get_fun"])

    FIELDS = [
        Field("Actual range start", "m", "{:.3f}", lambda d: d["range_start_m"]),
        Field("Actual range length", "m", "{:.3f}", lambda d: d["range_length_m"]),
        Field(
            "Actual range end",
            "m",
            "{:.3f}",
            lambda d: d["range_start_m"] + d["range_length_m"],
        ),
        Field("Step length", "mm", "{:.2f}", lambda d: d["step_length_m"] * 1e3),
        Field("Number of data points", "", "{}", lambda d: d["data_length"]),
        Field("Sweep rate", "Hz", "{:.0f}", lambda d: d["sweep_rate"]),
        Field("Stitch count", "", "{}", lambda d: d["stitch_count"]),
        Field("Depth LPF cutoff ratio", "", "{:.6f}", lambda d: d["depth_lowpass_cutoff_ratio"]),
    ]

    def __init__(self, parent):
        super().__init__(parent)

        grid = QGridLayout(self)
        grid.setContentsMargins(0, 0, 0, 0)
        self.setLayout(grid)

        self.field_widgets = []
        for i, field in enumerate(self.FIELDS):
            key_lbl = QLabel(field.label, self)
            val_lbl = QLabel("", self)
            unit_lbl = QLabel(field.unit, self)
            grid.addWidget(key_lbl, i, 0)
            grid.addWidget(val_lbl, i, 1, QtCore.Qt.AlignRight)
            grid.addWidget(unit_lbl, i, 2)
            self.field_widgets.append((key_lbl, val_lbl, unit_lbl))

        self.no_info_lbl = QLabel("No active or buffered session", self)
        grid.addWidget(self.no_info_lbl, i + 1, 0, 1, 3)

        self.update()

    def update(self, info=None):
        self.no_info_lbl.setVisible(not info)

        for field, widgets in zip(self.FIELDS, self.field_widgets):
            try:
                val = field.get_fun(info)
                text = field.fmt_str.format(val)
            except Exception:
                text = None

            for w in widgets:
                w.setVisible(bool(text))

            widgets[1].setText(str(text))


class QHLine(QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)


class QVLine(QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import ctypes
import ctypes.util
import enum
import os
import platform
from ctypes import POINTER, byref
from glob import glob


C_PVOID = ctypes.c_void_p
C_ULONG = ctypes.c_uint
C_DWORD = ctypes.c_uint
C_BOOL = ctypes.c_uint
C_STATUS = C_ULONG
C_HANDLE = C_PVOID
C_INT_ENUM = ctypes.c_ulong  # not part of original headers


class C_DeviceListInfoNode(ctypes.Structure):
    _fields_ = [
        ("Flags", C_ULONG),
        ("Type", C_ULONG),
        ("ID", C_ULONG),
        ("LocId", C_DWORD),
        ("SerialNumber", ctypes.c_char * 16),
        ("Description", ctypes.c_char * 64),
        ("ftHandle", C_HANDLE),
    ]


class C_VersionStruct(ctypes.Structure):
    _fields_ = [
        ("chipVersion", C_DWORD),
        ("dllVersion", C_DWORD),
    ]


OPEN_BY_SERIAL_NUMBER = 1
OPEN_BY_DESCRIPTION = 2
OPEN_BY_LOCATION = 4


class Status(enum.IntEnum):
    OK = 0
    INVALID_HANDLE = enum.auto()
    DEVICE_NOT_FOUND = enum.auto()
    DEVICE_NOT_OPENED = enum.auto()
    IO_ERROR = enum.auto()
    INSUFFICIENT_RESOURCES = enum.auto()
    INVALID_PARAMETER = enum.auto()
    INVALID_BAUD_RATE = enum.auto()
    DEVICE_NOT_OPENED_FOR_ERASE = enum.auto()
    DEVICE_NOT_OPENED_FOR_WRITE = enum.auto()
    FAILED_TO_WRITE_DEVICE = enum.auto()
    EEPROM_READ_FAILED = enum.auto()
    EEPROM_WRITE_FAILED = enum.auto()
    EEPROM_ERASE_FAILED = enum.auto()
    EEPROM_NOT_PRESENT = enum.auto()
    EEPROM_NOT_PROGRAMMED = enum.auto()
    INVALID_ARGS = enum.auto()
    NOT_SUPPORTED = enum.auto()
    OTHER_ERROR = enum.auto()
    DEVICE_LIST_NOT_READY = enum.auto()

    DEVICE_NOT_SUPPORTED = 1000
    CLK_NOT_SUPPORTED = enum.auto()
    VENDER_CMD_NOT_SUPPORTED = enum.auto()
    IS_NOT_SPI_MODE = enum.auto()
    IS_NOT_I2C_MODE = enum.auto()
    IS_NOT_SPI_SINGLE_MODE = enum.auto()
    IS_NOT_SPI_MULTI_MODE = enum.auto()
    WRONG_I2C_ADDR = enum.auto()
    INVAILD_FUNCTION = enum.auto()
    INVALID_POINTER = enum.auto()
    EXCEEDED_MAX_TRANSFER_SIZE = enum.auto()
    FAILED_TO_READ_DEVICE = enum.auto()
    I2C_NOT_SUPPORTED_IN_THIS_MODE = enum.auto()
    GPIO_NOT_SUPPORTED_IN_THIS_MODE = enum.auto()
    GPIO_EXCEEDED_MAX_PORTNUM = enum.auto()
    GPIO_WRITE_NOT_SUPPORTED = enum.auto()
    GPIO_PULLUP_INVALID_IN_INPUTMODE = enum.auto()
    GPIO_PULLDOWN_INVALID_IN_INPUTMODE = enum.auto()
    GPIO_OPENDRAIN_INVALID_IN_OUTPUTMODE = enum.auto()
    INTERRUPT_NOT_SUPPORTED = enum.auto()
    GPIO_INPUT_NOT_SUPPORTED = enum.auto()
    EVENT_NOT_SUPPORTED = enum.auto()
    FUN_NOT_SUPPORT = enum.auto()


class ClockRate(enum.IntEnum):
    SYS_CLK_60 = 0
    SYS_CLK_24 = enum.auto()
    SYS_CLK_48 = enum.auto()
    SYS_CLK_80 = enum.auto()


class SPIClock(enum.IntEnum):
    CLK_NONE = 0
    CLK_DIV_2 = enum.auto()
    CLK_DIV_4 = enum.auto()
    CLK_DIV_8 = enum.auto()
    CLK_DIV_16 = enum.auto()
    CLK_DIV_32 = enum.auto()
    CLK_DIV_64 = enum.auto()
    CLK_DIV_128 = enum.auto()
    CLK_DIV_256 = enum.auto()
    CLK_DIV_512 = enum.auto()


class SPIMode(enum.IntEnum):
    SPI_IO_NONE = 0
    SPI_IO_SINGLE = 1
    SPI_IO_DUAL = 2
    SPI_IO_QUAD = 4


class SPICPOL(enum.IntEnum):
    CLK_IDLE_LOW = 0
    CLK_IDLE_HIGH = 1


class SPICPHA(enum.IntEnum):
    CLK_LEADING = 0
    CLK_TRAILING = 1


class DrivingStrength(enum.IntEnum):
    DS_4MA = 0
    DS_8MA = enum.auto()
    DS_12MA = enum.auto()
    DS_16MA = enum.auto()


FUN_ARGTYPES = {
    "FT_CreateDeviceInfoList": [POINTER(C_DWORD)],
    "FT_GetDeviceInfoList": [POINTER(C_DeviceListInfoNode), POINTER(C_DWORD)],
    "FT4222_GetVersion": [C_HANDLE, POINTER(C_VersionStruct)],
    "FT_OpenEx": [C_PVOID, C_DWORD, POINTER(C_HANDLE)],
    "FT_Close": [C_HANDLE],
    "FT4222_SetClock": [C_HANDLE, C_INT_ENUM],
    "FT4222_GetClock": [C_HANDLE, POINTER(C_INT_ENUM)],
    "FT_SetTimeouts": [C_HANDLE, C_ULONG, C_ULONG],
    "FT4222_SetSuspendOut": [C_HANDLE, C_BOOL],
    "FT4222_SetWakeUpInterrupt": [C_HANDLE, C_BOOL],
    "FT4222_UnInitialize": [C_HANDLE],
    "FT4222_SPI_SetDrivingStrength": [C_HANDLE, C_INT_ENUM, C_INT_ENUM, C_INT_ENUM],
    "FT4222_SPIMaster_Init": [
        C_HANDLE,
        C_INT_ENUM,
        C_INT_ENUM,
        C_INT_ENUM,
        C_INT_ENUM,
        ctypes.c_uint8,
    ],
    "FT4222_SPIMaster_SingleRead": [
        C_HANDLE,
        POINTER(ctypes.c_uint8),
        ctypes.c_uint16,
        POINTER(ctypes.c_uint16),
        C_BOOL,
    ],
    "FT4222_SPIMaster_SingleWrite": [
        C_HANDLE,
        POINTER(ctypes.c_uint8),
        ctypes.c_uint16,
        POINTER(ctypes.c_uint16),
        C_BOOL,
    ],
    "FT4222_SPIMaster_SingleReadWrite": [
        C_HANDLE,
        POINTER(ctypes.c_uint8),
        POINTER(ctypes.c_uint8),
        ctypes.c_uint16,
        POINTER(ctypes.c_uint16),
        C_BOOL,
    ],
}


funs = None


def _load_dll():
    global funs

    if funs is not None:
        return

    system = platform.system().lower()
    bin_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data/libft4222")
    if system == "linux":
        machine_type = platform.machine()
        sub_dir = machine_type.lower()

        if sub_dir.startswith("arm"):
            sub_dir = sub_dir[:5]  # get 5 first letters: "armv?"
        try:
            lib_path = glob(os.path.join(bin_dir, sub_dir, "libft4222.so*"))[0]
            dll = ctypes.CDLL(lib_path)
        except (IndexError, OSError):
            raise OSError(f"Unsupported machine type: '{machine_type}'") from None

        funs = {name: getattr(dll, name) for name in FUN_ARGTYPES.keys()}
    elif system == "windows":
        try:
            ft_dll = ctypes.WinDLL(os.path.join(bin_dir, "amd64", "ftd2xx.dll"))
            ft4222_dll = ctypes.CDLL(os.path.join(bin_dir, "amd64", "LibFT4222.dll"))
        except OSError:
            ft_dll = ctypes.WinDLL(os.path.join(bin_dir, "i386", "ftd2xx.dll"))
            ft4222_dll = ctypes.CDLL(os.path.join(bin_dir, "i386", "LibFT4222.dll"))
        funs = {}
        for name in FUN_ARGTYPES.keys():
            prefix = name.split("_")[0].lower()
            dll = ft4222_dll if prefix == "ft4222" else ft_dll
            funs[name] = getattr(dll, name)
    else:
        raise RuntimeError("OS not supported")

    for name, argtypes in FUN_ARGTYPES.items():
        fun = funs[name]
        fun.restype = C_STATUS
        fun.argtypes = argtypes


def check_status(status):
    if status != Status.OK.value:
        try:
            status = Status(status)
        except ValueError:
            msg = "unknown"
            err = status
        else:
            msg = status.name
            err = status.value

        raise LibFT4222Error("Error {}: {}".format(err, msg))


def get_enum_val(x):
    return x.value if isinstance(x, enum.Enum) else x


class LibFT4222Error(Exception):
    pass


class Device:
    def __init__(self, handle=None):
        self.handle = handle
        _load_dll()

    def open_ex(self):
        _locationId = b"FT4222 A"
        self.handle = C_HANDLE()
        check_status(funs["FT_OpenEx"](_locationId, OPEN_BY_DESCRIPTION, byref(self.handle)))

    def close(self):
        check_status(funs["FT_Close"](self.handle))
        self.handle = None

    def get_version(self):
        _version = C_VersionStruct()
        check_status(funs["FT4222_GetVersion"](self.handle, byref(_version)))
        return int(_version.chipVersion), int(_version.dllVersion)

    def get_clock(self):
        _clock = C_INT_ENUM()
        check_status(funs["FT4222_GetClock"](self.handle, byref(_clock)))
        return ClockRate(int(_clock.value))

    def set_clock(self, clock):
        check_status(funs["FT4222_SetClock"](self.handle, get_enum_val(clock)))

    def set_timeouts(self, read, write):
        check_status(funs["FT_SetTimeouts"](self.handle, read, write))

    def set_suspend_out(self, enable):
        check_status(funs["FT4222_SetSuspendOut"](self.handle, enable))

    def set_wake_up_interrupt(self, enable):
        check_status(funs["FT4222_SetWakeUpInterrupt"](self.handle, enable))

    def spi_master_init(
        self,
        io_line=SPIMode.SPI_IO_SINGLE,
        clock=SPIClock.CLK_NONE,
        cpol=SPICPOL.CLK_IDLE_LOW,
        cpha=SPICPHA.CLK_LEADING,
        sso_map=1,
    ):
        status = funs["FT4222_SPIMaster_Init"](
            self.handle,
            get_enum_val(io_line),
            get_enum_val(clock),
            get_enum_val(cpol),
            get_enum_val(cpha),
            sso_map,
        )
        check_status(status)

    def spi_set_driving_strength(
        self, clock=DrivingStrength.DS_8MA, io=DrivingStrength.DS_8MA, sso=DrivingStrength.DS_8MA
    ):
        status = funs["FT4222_SPI_SetDrivingStrength"](
            self.handle,
            get_enum_val(clock),
            get_enum_val(io),
            get_enum_val(sso),
        )
        check_status(status)

    def spi_master_single_read(self, num_bytes, is_end_transaction=True):
        read_buffer = (ctypes.c_uint8 * num_bytes)()
        size_transferred = ctypes.c_uint16()
        status = funs["FT4222_SPIMaster_SingleRead"](
            self.handle,
            read_buffer,
            num_bytes,
            byref(size_transferred),
            is_end_transaction,
        )

        assert size_transferred.value == num_bytes
        check_status(status)

        return bytearray(read_buffer)

    def spi_master_single_write(self, data, is_end_transaction=True):
        num_bytes = len(data)
        write_buffer = (ctypes.c_uint8 * num_bytes)(*data)
        size_transferred = ctypes.c_uint16()
        status = funs["FT4222_SPIMaster_SingleWrite"](
            self.handle,
            write_buffer,
            num_bytes,
            byref(size_transferred),
            is_end_transaction,
        )

        assert size_transferred.value == num_bytes
        check_status(status)

    def spi_master_single_read_write(self, write_data, is_end_transaction=True):
        num_bytes = len(write_data)
        read_buffer = (ctypes.c_uint8 * num_bytes)()
        write_buffer = (ctypes.c_uint8 * num_bytes)(*write_data)
        size_transferred = ctypes.c_uint16()
        status = funs["FT4222_SPIMaster_SingleReadWrite"](
            self.handle,
            read_buffer,
            write_buffer,
            num_bytes,
            byref(size_transferred),
            is_end_transaction,
        )

        assert size_transferred.value == num_bytes
        check_status(status)

        return bytearray(read_buffer)


def print_devices():
    _load_dll()

    _num_devs = C_DWORD()
    status = funs["FT_CreateDeviceInfoList"](byref(_num_devs))
    check_status(status)

    _dev_info = (C_DeviceListInfoNode * _num_devs.value)()
    status = funs["FT_GetDeviceInfoList"](_dev_info, byref(_num_devs))
    check_status(status)

    for i, v in enumerate(_dev_info):
        print("{}:".format(i))
        print("  ", v.Type)
        print("  ", v.Description.decode() or "-")
        print("  ", v.LocId)


if __name__ == "__main__":
    print_devices()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from enum import Enum
from typing import Union


class _StrEnum(str, Enum):
    """Little middle-class that enables string equality checking and is an enum."""

    pass


class Protocol(_StrEnum):
    EXPLORATION = "exploration"
    MODULE = "module"
    STREAMING = "streaming"


class Link(_StrEnum):
    SOCKET = "socket"
    UART = "uart"
    SPI = "spi"


ProtocolArg = Union[str, Protocol]
LinkArg = Union[str, Link]



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Mapping, Optional

import attrs
import h5py
import numpy as np
import qtawesome as qta

from PySide6.QtWidgets import QLabel, QPushButton, QTabWidget, QVBoxLayout, QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    DetectorBackendPluginBase,
    DetectorPlotPluginBase,
    DetectorViewPluginBase,
)
from acconeer.exptool.app.new import (
    BUTTON_ICON_COLOR,
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GridGroupBox,
    HandledException,
    Message,
    MiscErrorView,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    PluginState,
    PluginStateMessage,
    VerticalGroupBox,
    is_task,
    pidgets,
)
from acconeer.exptool.app.new.ui.plugin_components import CollapsibleWidget, SensorConfigEditor

from ._configs import get_high_accuracy_detector_config
from ._detector import (
    DetailedStatus,
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorResult,
    PeakSortingMethod,
    ThresholdMethod,
    _load_algo_data,
)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_ids: list[int] = attrs.field(factory=lambda: [1])
    config: DetectorConfig = attrs.field(factory=DetectorConfig)
    context: DetectorContext = attrs.field(factory=DetectorContext)


class PluginPresetId(Enum):
    DEFAULT = auto()
    HIGH_ACCURACY = auto()


class BackendPlugin(DetectorBackendPluginBase[SharedState]):

    PLUGIN_PRESETS: Mapping[int, Callable[[], DetectorConfig]] = {
        PluginPresetId.DEFAULT.value: lambda: DetectorConfig(),
        PluginPresetId.HIGH_ACCURACY.value: lambda: get_high_accuracy_detector_config(),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)
        self._detector_instance: Optional[Detector] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])
        self.shared_state.context = DetectorContext.from_h5(file["context"])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState()
        self.broadcast(sync=True)

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            # Try to use the sensor ids from the last calibration
            if self.shared_state.context.sensor_ids:
                self.shared_state.sensor_ids = self.shared_state.context.sensor_ids

            for i in range(len(self.shared_state.sensor_ids)):
                if len(sensor_ids) > 0 and self.shared_state.sensor_ids[i] not in sensor_ids:
                    self.shared_state.sensor_ids[i] = sensor_ids[0]

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    @is_task
    def update_sensor_ids(self, *, sensor_ids: list[int]) -> None:
        self.shared_state.sensor_ids = sensor_ids
        self.broadcast()

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast(sync=True)

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())
        context_group = file.create_group("context")
        self.shared_state.context.to_h5(context_group)

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.context = context
        self.shared_state.sensor_ids = list(next(iter(record.session_config.groups)).keys())

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._detector_instance = Detector(
            client=self.client,
            sensor_ids=self.shared_state.sensor_ids,
            detector_config=self.shared_state.config,
            context=self.shared_state.context,
        )
        self._detector_instance.start(recorder)
        self.callback(
            GeneralMessage(
                name="setup",
                kwargs=dict(num_curves=len(self._detector_instance.processor_specs)),
                recipient="plot_plugin",
            )
        )

    def end_session(self) -> None:
        assert self._detector_instance
        self._detector_instance.stop()

    def get_next(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError

        assert self.client
        result = self._detector_instance.get_next()

        self._frame_count += 1

        self.callback(GeneralMessage(name="rate_stats", data=self.client._rate_stats))
        self.callback(GeneralMessage(name="frame_count", data=self._frame_count))
        self.callback(GeneralMessage(name="plot", data=result, recipient="plot_plugin"))

    @is_task
    def calibrate_detector(self) -> None:
        if self._started:
            raise RuntimeError

        if self.client is None:
            raise RuntimeError

        if not self.client.connected:
            raise RuntimeError

        self.callback(PluginStateMessage(state=PluginState.LOADED_BUSY))

        try:
            self._detector_instance = Detector(
                client=self.client,
                sensor_ids=self.shared_state.sensor_ids,
                detector_config=self.shared_state.config,
                context=None,
            )
            self._detector_instance.calibrate_detector()
        except Exception as exc:
            raise HandledException("Failed to calibrate detector") from exc
        finally:
            self.callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

        self.shared_state.context = self._detector_instance.context
        self.broadcast()


class PlotPlugin(DetectorPlotPluginBase):

    _DISTANCE_HISTORY_SPAN_MARGIN = 0.01

    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup_from_message(self, message: GeneralMessage) -> None:
        assert message.kwargs is not None
        self.setup(**message.kwargs)

    def update_from_message(self, message: GeneralMessage) -> None:
        self.update(message.data)  # type: ignore[arg-type]

    def setup(self, num_curves: int) -> None:
        self.num_curves = num_curves
        self.distance_history = [np.NaN] * 100

        win = self.plot_layout

        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.sweep_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.threshold_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_curves)]

        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweep_curves[0], "Sweep")
        sweep_plot_legend.addItem(self.threshold_curves[0], "Threshold")

        self.dist_history_plot = win.addPlot(row=1, col=0)
        self.dist_history_plot.setMenuEnabled(False)
        self.dist_history_plot.showGrid(x=True, y=True)
        self.dist_history_plot.addLegend()
        self.dist_history_plot.setLabel("left", "Estimated distance (m)")
        self.dist_history_plot.addItem(pg.PlotDataItem())
        self.dist_history_plot.setXRange(0, len(self.distance_history))

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.dist_history_curve = self.dist_history_plot.plot(**feat_kw)

        self.sweep_smooth_max = et.utils.SmoothMax()
        self.distance_hist_smooth_lim = et.utils.SmoothLimits(tau_decay=0.5)

    def update(self, multi_sensor_result: dict[int, DetectorResult]) -> None:
        # Get the first element as the plugin only supports single sensor operation.
        (result,) = list(multi_sensor_result.values())

        assert result.distances is not None

        self.distance_history.pop(0)
        if len(result.distances) != 0:
            self.distance_history.append(result.distances[0])
        else:
            self.distance_history.append(np.nan)

        max_val_in_plot = 0
        for idx, processor_result in enumerate(result.processor_results):
            assert processor_result.extra_result.used_threshold is not None
            assert processor_result.extra_result.distances_m is not None
            assert processor_result.extra_result.abs_sweep is not None

            abs_sweep = processor_result.extra_result.abs_sweep
            threshold = processor_result.extra_result.used_threshold
            distances_m = processor_result.extra_result.distances_m

            self.sweep_curves[idx].setData(distances_m, abs_sweep)
            self.threshold_curves[idx].setData(distances_m, threshold)

            max_val_in_subsweep = max(max(threshold), max(abs_sweep))
            if max_val_in_plot < max_val_in_subsweep:
                max_val_in_plot = max_val_in_subsweep

        self.sweep_plot.setYRange(0, self.sweep_smooth_max.update(max_val_in_plot))

        if np.any(~np.isnan(self.distance_history)):
            self.dist_history_curve.setData(self.distance_history)
            lims = self.distance_hist_smooth_lim.update(self.distance_history)
            lower_lim = max(0.0, lims[0] - self._DISTANCE_HISTORY_SPAN_MARGIN)
            upper_lim = lims[1] + self._DISTANCE_HISTORY_SPAN_MARGIN
            self.dist_history_plot.setYRange(lower_lim, upper_lim)
        else:
            self.dist_history_curve.setData([])


class ViewPlugin(DetectorViewPluginBase):

    sensor_config_editors: list[SensorConfigEditor]

    TEXT_MSG_MAP = {
        DetailedStatus.OK: "Ready to start.",
        DetailedStatus.END_LESSER_THAN_START: "'Range end' point must be greater than 'Range "
        + "start'.",
        DetailedStatus.CONTEXT_MISSING: "Run detector calibration.",
        DetailedStatus.CALIBRATION_MISSING: "Run detector calibration.",
        DetailedStatus.CONFIG_MISMATCH: (
            "Current configuration does not match the configuration "
            + "used during detector calibration. Run detector calibration."
        ),
        DetailedStatus.INVALID_DETECTOR_CONFIG_RANGE: (
            "Invalid detector config. Valid measurement"
            + " range is "
            + str(Detector.MIN_DIST_M)
            + "-"
            + str(Detector.MAX_DIST_M)
            + "m."
        ),
    }

    def __init__(self, app_model: AppModel, view_widget: QWidget) -> None:
        super().__init__(app_model=app_model, view_widget=view_widget)
        self._log = logging.getLogger(__name__)

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)
        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.start_button = QPushButton(
            qta.icon("fa5s.play-circle", color=BUTTON_ICON_COLOR),
            "Start measurement",
            self.sticky_widget,
        )
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(
            qta.icon("fa5s.stop-circle", color=BUTTON_ICON_COLOR),
            "Stop",
            self.sticky_widget,
        )
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        self.calibrate_detector_button = QPushButton(
            qta.icon("fa.circle", color=BUTTON_ICON_COLOR),
            "Calibrate detector",
            self.sticky_widget,
        )
        self.calibrate_detector_button.clicked.connect(self._on_calibrate_detector)

        self.defaults_button = QPushButton(
            qta.icon("mdi6.restore", color=BUTTON_ICON_COLOR),
            "Reset settings and calibrations",
            self.sticky_widget,
        )
        self.defaults_button.clicked.connect(self._send_defaults_request)

        self.message_box = QLabel(self.sticky_widget)
        self.message_box.setWordWrap(True)

        button_group = GridGroupBox("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.calibrate_detector_button, 1, 0, 1, -1)
        button_group.layout().addWidget(self.defaults_button, 3, 0, 1, -1)
        button_group.layout().addWidget(self.message_box, 4, 0, 1, -1)

        sticky_layout.addWidget(button_group)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)
        scrolly_layout.addWidget(self.misc_error_view)

        sensor_selection_group = VerticalGroupBox("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdParameterWidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_parameter_changed.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)
        scrolly_layout.addWidget(sensor_selection_group)

        self.config_editor = AttrsConfigEditor[DetectorConfig](
            title="Detector parameters",
            factory_mapping=self.get_pidget_mapping(),
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)
        scrolly_layout.addWidget(self.config_editor)

        self.sensor_config_editor_tabs = QTabWidget()
        self.sensor_config_editor_tabs.setStyleSheet("QTabWidget::pane { padding: 5px;}")
        self.sensor_config_editors = []
        self.range_labels = ["Close range", "Far range"]

        for label in self.range_labels:
            sensor_config_editor = SensorConfigEditor()
            sensor_config_editor.set_read_only(True)
            self.sensor_config_editors.append(sensor_config_editor)
            self.sensor_config_editor_tabs.addTab(sensor_config_editor, label)

        self.collapsible_widget = CollapsibleWidget(
            "Sensor config info", self.sensor_config_editor_tabs, self.scrolly_widget
        )
        scrolly_layout.addWidget(self.collapsible_widget)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "start_m": pidgets.FloatParameterWidgetFactory(
                name_label_text="Range start",
                suffix=" m",
                decimals=3,
            ),
            "end_m": pidgets.FloatParameterWidgetFactory(
                name_label_text="Range end",
                suffix=" m",
                decimals=3,
            ),
            "max_step_length": pidgets.OptionalIntParameterWidgetFactory(
                name_label_text="Max step length",
                checkbox_label_text="Set",
                limits=(1, None),
                init_set_value=12,
            ),
            "max_profile": pidgets.EnumParameterWidgetFactory(
                name_label_text="Max profile",
                enum_type=a121.Profile,
                label_mapping={
                    a121.Profile.PROFILE_1: "1 (shortest)",
                    a121.Profile.PROFILE_2: "2",
                    a121.Profile.PROFILE_3: "3",
                    a121.Profile.PROFILE_4: "4",
                    a121.Profile.PROFILE_5: "5 (longest)",
                },
            ),
            "num_frames_in_recorded_threshold": pidgets.IntParameterWidgetFactory(
                name_label_text="Num frames in rec. thr.",
                limits=(1, None),
            ),
            "threshold_method": pidgets.EnumParameterWidgetFactory(
                name_label_text="Threshold method",
                enum_type=ThresholdMethod,
                label_mapping={
                    ThresholdMethod.CFAR: "CFAR",
                    ThresholdMethod.FIXED: "Fixed",
                    ThresholdMethod.RECORDED: "Recorded",
                },
            ),
            "peaksorting_method": pidgets.EnumParameterWidgetFactory(
                name_label_text="Peak sorting method",
                enum_type=PeakSortingMethod,
                label_mapping={
                    PeakSortingMethod.CLOSEST: "Closest",
                    PeakSortingMethod.HIGHEST_RCS: "Highest RCS",
                },
            ),
            "fixed_threshold_value": pidgets.FloatParameterWidgetFactory(
                name_label_text="Fixed threshold value",
                decimals=1,
                limits=(0, None),
            ),
            "threshold_sensitivity": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Threshold sensitivity",
                decimals=2,
                limits=(0, 1),
                show_limit_values=False,
            ),
            "signal_quality": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Signal quality",
                decimals=1,
                limits=(-10, 35),
                show_limit_values=False,
                limit_texts=("Less power", "Higher quality"),
            ),
            "update_rate": pidgets.OptionalFloatParameterWidgetFactory(
                name_label_text="Update rate",
                checkbox_label_text="Set",
                limits=(1, None),
                init_set_value=20,
            ),
        }

    def on_backend_state_update(self, backend_plugin_state: Optional[SharedState]) -> None:
        if backend_plugin_state is not None and backend_plugin_state.config is not None:
            results = backend_plugin_state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

    def on_app_model_update(self, app_model: AppModel) -> None:
        state = app_model.backend_plugin_state
        if state is None:
            self.start_button.setEnabled(False)
            self.calibrate_detector_button.setEnabled(False)
            self.stop_button.setEnabled(False)
            self.defaults_button.setEnabled(False)

            self.config_editor.set_data(None)
            self.config_editor.setEnabled(False)
            self.sensor_id_pidget.set_selected_sensor(None, [])
            self.message_box.setText("")

            return

        assert isinstance(state, SharedState)

        try:
            session_config, _ = Detector._detector_to_session_config_and_processor_specs(
                state.config, state.sensor_ids
            )
        except Exception:
            pass  # Since the session config is read only there is no gain in handling this error
        else:
            tab_visible = [False, False]
            for group in session_config.groups:
                for _, sensor_config in group.items():
                    index = 0 if sensor_config.subsweeps[0].enable_loopback else 1
                    tab_visible[index] = True
                    sensor_config_editor = self.sensor_config_editors[index]
                    sensor_config_editor.set_data(sensor_config)
                    sensor_config_editor.sync()

            for i, sensor_config_editor in enumerate(self.sensor_config_editors):
                index = self.sensor_config_editor_tabs.indexOf(sensor_config_editor)
                self.sensor_config_editor_tabs.setTabVisible(index, tab_visible[i])

            self.collapsible_widget.widget = self.sensor_config_editor_tabs

        self.defaults_button.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)

        self.config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)
        self.config_editor.set_data(state.config)

        (sensor_id,) = state.sensor_ids
        self.sensor_id_pidget.set_selected_sensor(sensor_id, app_model.connected_sensors)
        self.sensor_id_pidget.setEnabled(app_model.plugin_state.is_steady)

        detector_status = Detector.get_detector_status(
            state.config, state.context, state.sensor_ids
        )

        self.message_box.setText(self.TEXT_MSG_MAP[detector_status.detector_state])

        ready_for_session = app_model.is_ready_for_session()

        self.calibrate_detector_button.setEnabled(ready_for_session)
        self.start_button.setEnabled(ready_for_session and detector_status.ready_to_start)
        self.stop_button.setEnabled(app_model.plugin_state == PluginState.LOADED_BUSY)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        self.app_model.put_backend_plugin_task("update_sensor_ids", {"sensor_ids": [sensor_id]})

    def _on_config_update(self, config: DetectorConfig) -> None:
        self.app_model.put_backend_plugin_task("update_config", {"config": config})

    def handle_message(self, message: GeneralMessage) -> None:
        if message.name == "sync":
            self._log.debug(f"{type(self).__name__} syncing")
            self.config_editor.sync()
        else:
            raise RuntimeError("Unknown message")

    def _on_calibrate_detector(self) -> None:
        self.app_model.put_backend_plugin_task(
            "calibrate_detector", on_error=self.app_model.emit_error
        )
        self.app_model.set_plugin_state(PluginState.LOADED_STARTING)

    def _send_defaults_request(self) -> None:
        self.app_model.put_backend_plugin_task("restore_defaults")


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPlugin:
        return ViewPlugin(app_model=app_model, view_widget=view_widget)

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPlugin:
        return PlotPlugin(app_model=app_model, plot_layout=plot_layout)


DISTANCE_DETECTOR_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="distance_detector",
    title="Distance detector",
    description="Easily measure distance to objects.",
    family=PluginFamily.DETECTOR,
    presets=[
        PluginPresetBase(name="Low power", preset_id=PluginPresetId.DEFAULT),
        PluginPresetBase(name="High accuracy", preset_id=PluginPresetId.HIGH_ACCURACY),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from ._bootloader_tool import BootloaderTool
from ._meta import (
    ACCONEER_XC120_BOOTLOADER_PID,
    ACCONEER_XC120_EXPLORATION_SERVER_PID,
    ACCONEER_XC120_EXPLORATION_SERVER_WINUSB_PID,
)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import logging
from functools import partial
from typing import Any, Mapping, Optional

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121._core import Criticality

from . import pidgets
from .range_help_view import RangeHelpView
from .types import PidgetFactoryMapping


log = logging.getLogger(__name__)


class SubsweepConfigEditor(QWidget):
    sig_update = Signal(object)

    _subsweep_config: Optional[a121.SubsweepConfig]

    _all_pidgets: list[pidgets.ParameterWidget]

    SPACING = 15
    PROFILE_LABEL_MAP = {
        a121.Profile.PROFILE_1: "1 (shortest)",
        a121.Profile.PROFILE_2: "2",
        a121.Profile.PROFILE_3: "3",
        a121.Profile.PROFILE_4: "4",
        a121.Profile.PROFILE_5: "5 (longest)",
    }
    PRF_LABEL_MAP = {
        a121.PRF.PRF_19_5_MHz: "19.5 MHz",
        a121.PRF.PRF_15_6_MHz: "15.6 MHz",
        a121.PRF.PRF_13_0_MHz: "13.0 MHz",
        a121.PRF.PRF_8_7_MHz: "8.7 MHz",
        a121.PRF.PRF_6_5_MHz: "6.5 MHz",
        a121.PRF.PRF_5_2_MHz: "5.2 MHz",
    }
    SUBSWEEP_CONFIG_FACTORIES: PidgetFactoryMapping = {
        "start_point": pidgets.IntParameterWidgetFactory(
            name_label_text="Start point:",
            name_label_tooltip=a121.SubsweepConfig.start_point.__doc__,
        ),
        "num_points": pidgets.IntParameterWidgetFactory(
            name_label_text="Number of points:",
            name_label_tooltip=a121.SubsweepConfig.num_points.__doc__,
            limits=(1, 4095),
        ),
        "step_length": pidgets.IntParameterWidgetFactory(
            name_label_text="Step length:",
            name_label_tooltip=a121.SubsweepConfig.step_length.__doc__,
            limits=(1, None),
        ),
        "hwaas": pidgets.IntParameterWidgetFactory(
            name_label_text="HWAAS:",
            name_label_tooltip=a121.SubsweepConfig.hwaas.__doc__,
            limits=(1, 511),
        ),
        "receiver_gain": pidgets.IntParameterWidgetFactory(
            name_label_text="Receiver gain:",
            name_label_tooltip=a121.SubsweepConfig.receiver_gain.__doc__,
            limits=(0, 23),
        ),
        "profile": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.Profile,
            name_label_text="Profile:",
            name_label_tooltip=a121.SubsweepConfig.profile.__doc__,  # type: ignore[arg-type]
            label_mapping=PROFILE_LABEL_MAP,
        ),
        "prf": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.PRF,
            name_label_text="PRF:",
            name_label_tooltip=a121.SubsweepConfig.prf.__doc__,  # type: ignore[arg-type]
            label_mapping=PRF_LABEL_MAP,
        ),
        "enable_tx": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Enable transmitter",
            name_label_tooltip=a121.SubsweepConfig.enable_tx.__doc__,
        ),
        "enable_loopback": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Enable loopback",
            name_label_tooltip=a121.SubsweepConfig.enable_loopback.__doc__,
        ),
        "phase_enhancement": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Phase enhancement",
            name_label_tooltip=a121.SubsweepConfig.phase_enhancement.__doc__,
        ),
    }

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)

        self._all_pidgets = []

        self._subsweep_config = None

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.range_help_view = RangeHelpView(self)
        self.layout().addWidget(self.range_help_view)

        self._subsweep_config_pidgets: Mapping[str, pidgets.ParameterWidget] = {}
        for aspect, factory in self.SUBSWEEP_CONFIG_FACTORIES.items():
            pidget = factory.create(self)
            self.layout().addWidget(pidget)

            pidget.sig_parameter_changed.connect(
                partial(self._update_subsweep_config_aspect, aspect)
            )

            self._all_pidgets.append(pidget)
            self._subsweep_config_pidgets[aspect] = pidget

    def sync(self) -> None:
        self._update_ui()

    def _update_ui(self) -> None:
        if self._subsweep_config is None:
            log.debug("could not update ui as SubsweepConfig is None")
            return

        self._subsweep_config_pidgets["start_point"].set_parameter(
            self._subsweep_config.start_point
        )
        self._subsweep_config_pidgets["num_points"].set_parameter(self._subsweep_config.num_points)
        self._subsweep_config_pidgets["step_length"].set_parameter(
            self._subsweep_config.step_length
        )
        self._subsweep_config_pidgets["profile"].set_parameter(self._subsweep_config.profile)
        self._subsweep_config_pidgets["hwaas"].set_parameter(self._subsweep_config.hwaas)
        self._subsweep_config_pidgets["receiver_gain"].set_parameter(
            self._subsweep_config.receiver_gain
        )
        self._subsweep_config_pidgets["enable_tx"].set_parameter(self._subsweep_config.enable_tx)
        self._subsweep_config_pidgets["enable_loopback"].set_parameter(
            self._subsweep_config.enable_loopback
        )
        self._subsweep_config_pidgets["phase_enhancement"].set_parameter(
            self._subsweep_config.phase_enhancement
        )
        self._subsweep_config_pidgets["prf"].set_parameter(self._subsweep_config.prf)

    def set_data(self, subsweep_config: Optional[a121.SubsweepConfig]) -> None:
        self.range_help_view.update(subsweep_config)
        self._subsweep_config = subsweep_config

    def set_read_only(self, read_only: bool) -> None:
        for pidget in self._all_pidgets:
            pidget.setEnabled(not read_only)

    def _update_subsweep_config_aspect(self, aspect: str, value: Any) -> None:
        if self._subsweep_config is None:
            raise TypeError("SubsweepConfig is None")

        try:
            setattr(self._subsweep_config, aspect, value)
        except Exception as e:
            self._subsweep_config_pidgets[aspect].set_note_text(e.args[0], Criticality.ERROR)

        self._broadcast()

    def handle_validation_results(
        self, results: list[a121.ValidationResult]
    ) -> list[a121.ValidationResult]:
        for pidget in self._all_pidgets:
            pidget.set_note_text("")

        unhandled_results: list[a121.ValidationResult] = []

        for result in results:
            if not self._handle_validation_result(result):
                unhandled_results.append(result)

        return unhandled_results

    def _handle_validation_result(self, result: a121.ValidationResult) -> bool:
        if result.aspect is None or self._subsweep_config is None:
            return False

        result_handled = False

        if result.source == self._subsweep_config:
            self._subsweep_config_pidgets[result.aspect].set_note_text(
                result.message, result.criticality
            )
            result_handled = True

        return result_handled

    def _broadcast(self) -> None:
        self.sig_update.emit(self._subsweep_config)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

ACCONEER_XM_CP2105_MODULE_VID = 0x10C4
ACCONEER_XM_CP2105_MODULE_PID = 0xEA70

STM32L431_ID = 0x0435

DEVICE_NAME_TO_ID = {
    "Accconeer XE125": STM32L431_ID,
    "XE125": STM32L431_ID,
    "XM125": STM32L431_ID,
}
